# -*- coding: utf-8 -*-
"""
hyperframe/frame
~~~~~~~~~~~~~~~~

Defines framing logic for HTTP/2. Provides both classes to represent framed
data and logic for aiding the connection when it comes to reading from the
socket.
"""
import h2.errors
import sys
import struct
from typing import Any, Optional,  List, Iterable
from hedra.core.engines.types.http2.stream import Stream
from hedra.core.engines.types.http2.events.remote_settings_changed_event import RemoteSettingsChanged
from hedra.core.engines.types.http2.events.settings_acknowledged_event import SettingsAcknowledged
from hedra.core.engines.types.http2.streams.stream_settings import SettingCodes
from hedra.core.engines.types.http2.events.stream_reset import StreamReset
from hedra.core.engines.types.http2.streams.stream_closed_by import StreamClosedBy
from hedra.core.engines.types.http2.events.deferred_headers_event import DeferredHeaders
from hedra.core.engines.types.http2.events.connection_terminated_event import ConnectionTerminated
from hedra.core.engines.types.http2.errors.exceptions import StreamClosedError
from hedra.core.engines.types.http2.events.data_received_event import DataReceived
from hedra.core.engines.types.http2.events.stream_ended_event import StreamEnded
from hedra.core.engines.types.http2.errors.exceptions import StreamError
from hedra.core.engines.types.http2.errors.types import ErrorCodes
from hedra.core.engines.types.http2.events.window_updated_event import WindowUpdated

from .attributes import (
    Flag,
    Flags,
    _STRUCT_HBBBL,
    _STRUCT_H,
    _STRUCT_B,
    _STRUCT_LL,
    _STRUCT_LB,
    _STRUCT_L,
    _STRUCT_HL
)
from .utils import raw_data_repr


class Frame:
    __slots__ = (
        'stream_id', 
        'flags',
        'body_len',
        'flags',
        'type',
        'frame_type',
        'data',
        'settings',
        'origin',
        'field',
        'error_code',
        'pad_length',
        'last_stream_id',
        'additional_data',
        'depends_on',
        'stream_weight',
        'exclusive',
        'opaque_data',
        'promised_stream_id',
        'window_increment',
        'flag_byte',
        'defined_flags'
    )
    
    FRAMES = {}
    """
    The base class for all HTTP/2 frames.
    """
    #: The flags defined on this type of frame.

    # If 'has-stream', the frame's stream_id must be non-zero. If 'no-stream',
    # it must be zero. If 'either', it's not checked.
    stream_association: Optional[str] = None
    frame_types = {
        0xA: sys.intern('ALTSVC'),
        0x09: sys.intern('CONTINUATION'),
        0x0: sys.intern('DATA'),
        0x07: sys.intern('GOAWAY'),
        0x01: sys.intern('HEADERS'),
        0x06: sys.intern('PING'),
        0x02: sys.intern('PRIORITY'),
        0x05: sys.intern('PUSHPROMISE'),
        0x03: sys.intern('RESET'),
        0x04: sys.intern('SETTINGS'),
        0x08: sys.intern('WINDOWUPDATE')
    }

    def __init__(self, stream_id: int, frame_type: int, flags: Iterable[str] = (), parsed_flag_byte: int = 0, **kwargs: Any) -> None:
        #: The stream identifier for the stream this frame was received on.
        #: Set to 0 for frames sent on the connection (stream-id 0).
        self.stream_id = stream_id
        self.type = frame_type
        self.frame_type = self.frame_types.get(self.type)
        self.field = b''
        self.data = b''
        self.settings = {}
        self.origin = b''
        self.error_code = 0
        self.pad_length = 0
        self.last_stream_id = 0
        self.additional_data = 0
        self.depends_on = 0x0
        self.stream_weight = 0x0
        self.exclusive = False
        self.opaque_data = b''
        self.promised_stream_id = 0
        self.window_increment = 0
        self.flag_byte = 0x0
        self.defined_flags: List[Flag] = []

        #: The flags set for this frame.
        self.flags = Flags(self.defined_flags)

        #: The frame length, excluding the nine-byte header.
        self.body_len = 0

        for flag in flags:
            self.flags.add(flag)

        if self.type == 0xA:
            # ALTSVC
            self.origin = kwargs.get('origin', b'')
            self.field = kwargs.get('fields', b'')

        elif self.type == 0x09:
            # CONTINUATION

            self.defined_flags = [
                Flag('END_HEADERS', 0x04)
            ]

            self.data = kwargs.get('data')

        elif self.type == 0x0:
            # DATA

            self.defined_flags = [
                Flag('END_STREAM', 0x01),
                Flag('PADDED', 0x08),
            ]

            self.pad_length = kwargs.get('pad_length', 0)
            self.data = kwargs.get('data', b'')

        elif self.type == 0x07:
            # GOAWAY
            self.last_stream_id = kwargs.get('last_stream_id', 0)
            self.additional_data = kwargs.get('additional_data', b'')
            self.error_code = kwargs.get('error_code', 0)

        elif self.type == 0x01:
            # HEADERS
            self.defined_flags = [
                Flag('END_STREAM', 0x01),
                Flag('END_HEADERS', 0x04),
                Flag('PADDED', 0x08),
                Flag('PRIORITY', 0x20),
            ]

            self.data = kwargs.get('data', b'')
            self.pad_length = kwargs.get('pad_length', 0)
            self.depends_on = kwargs.get('depends_on', 0x0)
            self.stream_weight = kwargs.get('stream_weight', 0x0)
            self.exclusive = kwargs.get('exclusive', False)

        elif self.type == 0x06:
            # PING

            self.defined_flags = [
                Flag('ACK', 0x01)
            ]

            self.opaque_data = kwargs.get('opaque_data', b'')

        elif self.type == 0x02:
            # PRIORITY
            self.depends_on = kwargs.get('depends_on', 0x0)
            self.stream_weight = kwargs.get('stream_weight', 0x0)
            self.exclusive = kwargs.get('exclusive', False)

        elif self.type == 0x05:
            # PUSH PROMISE
            self.defined_flags = [
                Flag('END_HEADERS', 0x04),
                Flag('PADDED', 0x08)
            ]

            self.promised_stream_id = kwargs.get('promised_stream_id', 0)
            self.pad_length = kwargs.get('pad_length', 0)
            self.data = kwargs.get('data', b'')

        elif self.type == 0x03:
            # RESET
            self.error_code = kwargs.get('error_code', 0)

        elif self.type == 0x04:
            # SETTINGS
            self.defined_flags = [
                Flag('ACK', 0x01)
            ]

            self.settings = kwargs.get('settings', {})

        elif self.type == 0x08:
            # WINDOW UPDATE
            self.window_increment = kwargs.get('window_increment', 0)

        else:
            # EXTENSION
            self.flag_byte = kwargs.get('flag_byte', 0x0)

        for flag, flag_bit in self.defined_flags:
            if parsed_flag_byte & flag_bit:
                self.flags.add(flag)

    def __repr__(self) -> str:
        body_repr = self._body_repr(),
        return f"{type(self).__name__}(stream_id={self.stream_id}, flags={repr(self.flags)}): {body_repr}"

    def _body_repr(self) -> str:
        # More specific implementation may be provided by subclasses of Frame.
        # This fallback shows the serialized (and truncated) body content.
        return raw_data_repr(self.serialize())

    @property
    def flow_controlled_length(self) -> int:
        """
        The length of the frame that needs to be accounted for when considering
        flow control.
        """
        padding_len = 0
        if 'PADDED' in self.flags:
            # Account for extra 1-byte padding length field, which is still
            # present if possibly zero-valued.
            padding_len = self.pad_length + 1
        return len(self.data) + padding_len

    def parse_flags(self, flag_byte: int) -> Flags:

        for flag, flag_bit in self.defined_flags:
            if flag_byte & flag_bit:
                self.flags.add(flag)

        return self.flags

    def serialize(self) -> bytes:
        """
        Convert a frame into a bytestring, representing the serialized form of
        the frame.
        """

        body = b''
        flags = 0

        if self.type == 0xA:
            # ALTSVC
            origin_len = _STRUCT_H.pack(len(self.origin))    
            body = origin_len + self.origin + self.field

        elif self.type == 0x09:
            # CONTINUATION

            body = self.data

        elif self.type == 0x0:
            # DATA

            padding_data = b''
            if 'PADDED' in self.flags:  # type: ignore
                padding_data = _STRUCT_B.pack(self.pad_length)


            padding = b'\0' * self.pad_length
            body = padding_data + self.data + padding

        elif self.type == 0x07:
            # GOAWAY

            self.data = _STRUCT_LL.pack(
                self.last_stream_id & 0x7FFFFFFF,
                self.error_code
            )
            
            body = self.data + self.additional_data

        elif self.type == 0x01:
            # HEADERS

            padding_data = b''
            if 'PADDED' in self.flags:  # type: ignore
                padding_data = _STRUCT_B.pack(self.pad_length)

            padding = b'\0' * self.pad_length

            if 'PRIORITY' in self.flags:
                priority_data = _STRUCT_LB.pack(
                    self.depends_on + (0x80000000 if self.exclusive else 0),
                    self.stream_weight
                )
            else:
                priority_data = b''

            body = padding_data + priority_data + self.data + padding

        elif self.type == 0x06:
            # PING
            
            body = self.opaque_data
            body += b'\x00' * (8 - len(body))

        elif self.type == 0x02:
            # PRIORITY
            body = _STRUCT_LB.pack(
                self.depends_on + (0x80000000 if self.exclusive else 0),
                self.stream_weight
            )

        elif self.type == 0x05:
            # PUSH PROMISE

            padding_data = b''
            if 'PADDED' in self.flags:  # type: ignore
                padding_data = _STRUCT_B.pack(self.pad_length)

            padding = b'\0' * self.pad_length
            promise_data = _STRUCT_L.pack(self.promised_stream_id)

            body = padding_data + promise_data + self.data + padding

        elif self.type == 0x03:
            # RESET
            body = _STRUCT_L.pack(self.error_code)

        elif self.type == 0x04:
            # SETTINGS
            for setting, value in self.settings.items():
                body += _STRUCT_HL.pack(setting & 0xFF, value)

        elif self.type == 0x08:
            # WINDOW UPDATE
            body = _STRUCT_L.pack(self.window_increment & 0x7FFFFFFF)

        else:
            # EXTENSION
            flags = self.flag_byte

        self.body_len = len(body)


        for flag, flag_bit in self.defined_flags:
            if flag in self.flags:
                flags |= flag_bit

        header = _STRUCT_HBBBL.pack(
            (self.body_len >> 8) & 0xFFFF,  # Length spread over top 24 bits
            self.body_len & 0xFF,
            self.type,
            flags,
            self.stream_id & 0x7FFFFFFF  # Stream ID is 32 bits.
        )

        return header + body

    def get_events_and_frames(self, stream: Stream, connection):

        if self.type == 0x0:
            # DATA
            end_stream = 'END_STREAM' in self.flags
            flow_controlled_length = self.flow_controlled_length
            frame_data = self.data

            frames = []
            data_events = []
            connection._inbound_flow_control_window_manager.window_consumed(
                flow_controlled_length
            )

            try:
                
                stream.inbound.window_consumed(flow_controlled_length)
        
                event = DataReceived()
                event.stream_id = stream.stream_id

                data_events.append(event)

                if end_stream:
                    event = StreamEnded()
                    event.stream_id = stream.stream_id
                    data_events[0].stream_ended = event
                    data_events.append(event)

                data_events[0].data = frame_data
                data_events[0].flow_controlled_length = flow_controlled_length
                return frames, data_events

            except StreamClosedError as e:
                # This stream is either marked as CLOSED or already gone from our
                # internal state.
                
                conn_manager = connection._inbound_flow_control_window_manager
                conn_increment = conn_manager.process_bytes(
                    flow_controlled_length
                )

                if conn_increment:
                    f = Frame(0)
                    f.window_increment = conn_increment
                    frames.append(f)

                f = Frame(e.stream_id, 0x03)
                f.error_code = e.error_code
                frames.append(f)
        
                return frames, data_events + e._events

        elif self.type == 0x07:
            # GOAWAY
            self._data_to_send = b''

            new_event = ConnectionTerminated()
            new_event.error_code = h2.errors._error_code_from_int(self.error_code)
            new_event.last_stream_id = self.last_stream_id
            
            if self.additional_data:
                new_event.additional_data = self.additional_data

            return [], [new_event]

        elif self.type == 0x01:
            # HEADERS

            stream_events = []
            deferred_headers = DeferredHeaders(
                stream.encoder,
                self,
                connection._h2_state.config.header_encoding
            )

            stream_events.append(deferred_headers)

            if deferred_headers.end_stream:
                event = StreamEnded()
                event.stream_id = stream.stream_id
                
                stream_events[0].stream_ended = event
                stream_events.append(event)

            return [], stream_events

        elif self.type == 0x03:
            # RESET

            stream.closed_by = StreamClosedBy.RECV_RST_STREAM
            reset_event = StreamReset()
            reset_event.stream_id = stream.stream_id
            reset_event[0].error_code = h2.errors._error_code_from_int(self.error_code)

            return [], [reset_event]

        elif self.type == 0x04:
            # SETTINGS
            stream_events = []
            if 'ACK' in self.flags:

                changes = connection.local_settings.acknowledge()
        
                initial_window_size_change = changes.get(SettingCodes.INITIAL_WINDOW_SIZE)
                max_header_list_size_change = changes.get(SettingCodes.MAX_HEADER_LIST_SIZE)
                max_frame_size_change = changes.get(SettingCodes.MAX_FRAME_SIZE)
                header_table_size_change =changes.get(SettingCodes.HEADER_TABLE_SIZE)

                if initial_window_size_change is not None:

                    window_delta = initial_window_size_change.new_value - initial_window_size_change.original_value
                    
                    new_max_window_size = stream.inbound.max_window_size + window_delta
                    stream.inbound.window_opened(window_delta)
                    stream.inbound.max_window_size = new_max_window_size

                if max_header_list_size_change is not None:
                    connection._decoder.max_header_list_size = max_header_list_size_change.new_value

                if max_frame_size_change is not None:
                    stream.max_outbound_frame_size =  max_frame_size_change.new_value

                if header_table_size_change:
                    # This is safe across all hpack versions: some versions just won't
                    # respect it.
                    connection._decoder.max_allowed_table_size = header_table_size_change.new_value

                ack_event = SettingsAcknowledged()
                ack_event.changed_settings = changes
                stream_events.append(ack_event)
                return [], stream_events

            # Add the new settings.
            connection.remote_settings.update(self.settings)
            stream_events.append(
                RemoteSettingsChanged.from_settings(
                    connection.remote_settings, self.settings
                )
            )

            changes = connection.remote_settings.acknowledge()
            initial_window_size_change = changes.get(SettingCodes.INITIAL_WINDOW_SIZE)
            header_table_size_change = changes.get(SettingCodes.HEADER_TABLE_SIZE)
            max_frame_size_change = changes.get(SettingCodes.MAX_FRAME_SIZE)
        
            if initial_window_size_change:
                stream.current_outbound_window_size = connection._guard_increment_window(
                    stream.current_outbound_window_size,
                    initial_window_size_change.new_value - initial_window_size_change.original_value
                )

            # HEADER_TABLE_SIZE changes by the remote part affect our encoder: cf.
            # RFC 7540 Section 6.5.2.
            if  header_table_size_change:
                connection._encoder.header_table_size = header_table_size_change.new_value

            if max_frame_size_change:
                stream.max_outbound_frame_size = max_frame_size_change.new_value

            frames = Frame(0, 0x04)
            frames.flags.add('ACK')

            return [frames], stream_events

        elif self.type == 0x08:
            # WINDOW UPDATE
            stream_events = []
            frames = []
            increment = self.window_increment
            if self.stream_id:
                try:

                    
                    event = WindowUpdated()
                    event.stream_id = stream.stream_id

                    # If we encounter a problem with incrementing the flow control window,
                    # this should be treated as a *stream* error, not a *connection* error.
                    # That means we need to catch the error and forcibly close the stream.
                    event.delta = increment

                    try:
                        connection.outbound_flow_control_window = connection._guard_increment_window(
                            connection.outbound_flow_control_window,
                            increment
                        )
                    except StreamError:
                        # Ok, this is bad. We're going to need to perform a local
                        # reset.

                        event = StreamReset()
                        event.stream_id = stream.stream_id
                        event.error_code = ErrorCodes.FLOW_CONTROL_ERROR
                        event.remote_reset = False

                        stream.closed_by = ErrorCodes.FLOW_CONTROL_ERROR    
                        
                        rsf = Frame(stream.stream_id, 0x03)
                        rsf.error_code = ErrorCodes.FLOW_CONTROL_ERROR

                        frames = [rsf]

                    stream_events.append(event)
                except Exception:
                    return [], stream_events
            else:
                connection.outbound_flow_control_window = connection._guard_increment_window(
                    connection.outbound_flow_control_window,
                    increment
                )
                # FIXME: Should we split this into one event per active stream?
                window_updated_event = WindowUpdated()
                window_updated_event.stream_id = 0
                window_updated_event.delta = increment
                stream_events.append(window_updated_event)
                frames = []

            return frames, stream_events

        return None, None



from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    CSVConfig
)


class SubmitCSVResultsStage(Submit):
    config=CSVConfig(
        events_filepath='./events.json',
        metrics_filepath='./metrics.json'
    )


import time
import asyncio


class Job:

    def __init__(self, job_id, pipeline) -> None:

        self.job_id = job_id
        self.start = None
        self.elapsed = 0
        self.timeout = pipeline.timeout
        
        self.task = None
        self.registered_leaders = []
        self.excluded_leaders = []
        self.pipeline = pipeline
        self.results = None

        self.status = None

    async def run(self):
        self.task = asyncio.create_task(
            self.pipeline.run(job_id=self.job_id)
        )

    async def set_job_status(self, status):
        self.status = status

    async def start_timer(self):
        self.start = time.time()

    async def update_time_elapsed(self):
        self.elapsed = time.time() - self.start

    async def check_if_deadline_exceeded(self):
        return self.elapsed > self.timeout

    async def attach_leader_to_job(self, leader_address):
        self.registered_leaders += [leader_address]
        self.registered_leaders = list(set(self.registered_leaders))

    async def attach_leader_to_excluded(self, leader_address):
        self.excluded_leaders += [leader_address]
        self.excluded_leaders = list(set(self.excluded_leaders))

    async def remove_leader(self, leader_address):
        if leader_address in self.registered_leaders:
            self.registered_leaders.remove(leader_address)

    async def check_if_completed(self):
        return len(self.registered_leaders) < 1


    async def get_task(self):
        if self.results is None:
            self.results = await self.task
        
        return self.results






import asyncio
from sys import path
import traceback
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.types.stage_states import StageStates
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.graphs.transitions.exceptions import (
    StageExecutionError,
    StageTimeoutError
)
from hedra.logging import HedraLogger


async def optimize_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - Executing transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Executing transition from {current_stage.name} to {next_stage.name}')

    execute_stages = current_stage.context.stages.get(StageTypes.EXECUTE).items()
    paths = current_stage.context.paths.get(current_stage.name)

    visited = current_stage.context.visited

    execute_stages = {
        stage_name: stage for stage_name, stage in execute_stages if stage_name in paths and stage_name not in visited
    }

    optimization_candidates = {}

    valid_states = [
        StageStates.INITIALIZED,
        StageStates.SETUP,
    ]

    for stage_name, stage in execute_stages.items():

        if stage_name in current_stage.context.paths and stage.state in valid_states:
            stage.state = StageStates.OPTIMIZING
            optimization_candidates[stage_name] = stage

    current_stage.generation_optimization_candidates = len(optimization_candidates)


    if current_stage.timeout:
        optimization_results = await asyncio.wait_for(current_stage.run(optimization_candidates), timeout=current_stage.timeout)

    else:
        optimization_results = await current_stage.run(optimization_candidates)

    next_stage.context.optimized_params = optimization_results
    next_stage.state = StageStates.OPTIMIZED

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - Completed transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Completed transition from {current_stage.name} to {next_stage.name}')

    next_stage.context = current_stage.context


async def optimize_to_execute_transition(current_stage: Stage, next_stage: Stage):

    try:

        await optimize_transition(current_stage, next_stage)
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR
    
    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    
    current_stage = None

    return None, StageTypes.EXECUTE


async def optimize_to_checkpoint_transition(current_stage: Stage, next_stage: Stage):

    try:

        next_stage.previous_stage = current_stage.name

        await optimize_transition(current_stage, next_stage)

        next_stage.data = next_stage.context.optimized_params
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR
    
    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    current_stage = None

    return None, StageTypes.CHECKPOINT


async def optimize_to_wait_transition(current_stage: Stage, next_stage: Stage):

    try:

        next_stage.previous_stage = current_stage.name
        
        await optimize_transition(current_stage, next_stage)

        next_stage.data = next_stage.context.optimized_params
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR
    
    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    current_stage = None

    return None, StageTypes.WAIT



import asyncio
import uuid
import time
import threading
import os
from typing import Any, Dict
from hedra.core.personas.batching.param_type import ParamType
from hedra.logging import HedraLogger
from hedra.tools.data_structures import AsyncList
from .algorithms import get_algorithm


class Optimizer:

    def __init__(self, config: Dict[str, Any]) -> None:

        self.logger = HedraLogger()
        self.logger.initialize()

        self.optimizer_id = str(uuid.uuid4())
        self.thread_id = threading.current_thread().ident
        self.process_id = os.getpid()
        
        self.graph_name = config.get('graph_name')
        self.graph_id = config.get('graph_id')
        self.source_stage_name = config.get('source_stage_name')
        self.source_stage_id = config.get('source_stage_id')

        self.metadata_string = f'Graph - {self.graph_name}:{self.graph_id} - thread:{self.thread_id} - process:{self.process_id} - Stage: {self.source_stage_name}:{self.source_stage_id} - Optimizer: {self.optimizer_id} - '

        self.stage_name = config.get('stage_name')
        self.actions = AsyncList()

        self.algorithm_type = config.get('algorithm', 'shg')
        persona = config.get('persona')

        self._optimization_time_limit = config.get('time_limit', 60)

        self.algorithm = get_algorithm(
            self.algorithm_type,
            {
                **config,
                'persona': persona
            }
        )

        self._current_iter = 0
        self.optimized_results = {}
        self._max_aps = 0
        self._event_loop: asyncio.AbstractEventLoop = None
        

    def optimize(self, loop):
        self._event_loop = loop

        results = None

        self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Starting optimization')
        self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Optimization config: Time Limit - {self._optimization_time_limit}')
        self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Optimization config: Algorithm - {self.algorithm_type}')
        self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Optimization config: Batch Time - {self.algorithm.batch_time}')
        self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Optimization config: Max Iter - {self.algorithm.max_iter}')

        start = time.time()
        
        results = self.algorithm.optimize(self._run_optimize)
        optimized_batch_size = int(results.x[0])

        self.total_optimization_time = time.time() - start

        self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Optimization took - {round(self.total_optimization_time, 2)} - seconds')
        self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Optimization - max actions per second - {self._max_aps}')
        self.logger.filesystem.sync['hedra.optimize'].info(f'{self.metadata_string} - Optimization - optimized batch size - {optimized_batch_size}')

        self.algorithm.persona.total_time = self.algorithm.persona_total_time

    
        self.optimized_results = {
            'optimized_batch_size': optimized_batch_size,
            'optimization_iters': self.algorithm.max_iter,
            'optimization_iter_duation': self.algorithm.batch_time,
            'optimization_total_time': self.total_optimization_time,
            'optimization_max_aps': self._max_aps
        }

        return self.optimized_results

    async def _optimize(self, *params):

        self.current_params = {}
        for idx, param in enumerate(params):
            param_name = self.algorithm.param_names[idx]
            self.current_params[param_name] = param

        if self._current_iter < self.algorithm.max_iter:

            await self.logger.filesystem.aio['hedra.optimize'].debug(f'{self.metadata_string} - Optimizer iteration - {self._current_iter}')

            self.algorithm.update_params()

            await self.logger.filesystem.aio['hedra.optimize'].debug(f'{self.metadata_string} - Optimizer iteration - {self._current_iter} - Batch Size - {self.algorithm.persona.batch.size}')
            await self.logger.filesystem.aio['hedra.optimize'].debug(f'{self.metadata_string} - Optimizer iteration - {self._current_iter} - Batch Interval - {self.algorithm.persona.batch.interval}')
            await self.logger.filesystem.aio['hedra.optimize'].debug(f'{self.metadata_string} - Optimizer iteration - {self._current_iter} - Batch Gradient - {self.algorithm.persona.batch.gradient}')

            try:
                completed = await asyncio.wait_for(
                    self.algorithm.persona.execute(), 
                    timeout=self.algorithm.batch_time * 2
                )

                completed_count = len([complete for complete in completed if complete.error is None])
            except asyncio.TimeoutError:
                completed_count = 1

            elapsed = self.algorithm.persona.end - self.algorithm.persona.start

            await self.logger.filesystem.aio['hedra.optimize'].debug(f'{self.metadata_string} - Optimizer iteration - {self._current_iter} - took - {round(elapsed, 2)} - seconds')
       
            if completed_count < 1:
                completed_count = elapsed

            await self.logger.filesystem.aio['hedra.optimize'].debug(f'{self.metadata_string} - Optimizer iteration - {self._current_iter} - actions per second - {round(completed/elapsed)}')
            await self.logger.filesystem.aio['hedra.optimize'].debug(f'{self.metadata_string} - Optimizer iteration - {self._current_iter} - Inverted APS score- {elapsed/completed}')

            return elapsed/completed_count

        return 0

    def _run_optimize(self, xargs):

        for idx, param in enumerate(xargs):
            param_name = self.algorithm.param_names[idx]
            param = self.algorithm.param_values.get(param_name, {})
            param_type = param.get('type')

            if param_type == ParamType.INTEGER:
                xargs[idx] = int(xargs[idx])

        inverse_actions_per_second = self._event_loop.run_until_complete(
            self._optimize(*xargs)
        )

        self._current_iter += 1


        return inverse_actions_per_second


import asyncio
import inspect
from typing import List, Union
from hedra.core.graphs.events import Event
from hedra.core.graphs.hooks.registry.registrar import registrar
from hedra.core.graphs.hooks.hook_types.internal import Internal
from hedra.core.graphs.hooks.registry.registry_types import (
    EventHook,
    ContextHook,
    TeardownHook
)
from hedra.core.graphs.hooks.registry.registry_types.hook import Hook
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.graphs.stages.base.stage import Stage


class Teardown(Stage):
    stage_type=StageTypes.TEARDOWN

    def __init__(self) -> None:
        super().__init__()
        self.actions = []
        self.accepted_hook_types = [ HookType.EVENT, HookType.CONTEXT ]

    @Internal()
    async def run(self):

        events: List[Union[EventHook, Event]] = [event for event in self.hooks[HookType.EVENT]]
        pre_events: List[EventHook] = [
            event for event in events if isinstance(event, EventHook) and event.pre
        ]
        
        if len(pre_events) > 0:
            pre_event_names = ", ".join([
                event.shortname for event in pre_events
            ])

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing PRE events - {pre_event_names}')
            await asyncio.wait([
                asyncio.create_task(event.call()) for event in pre_events
            ], timeout=self.stage_timeout)

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing PRE events - {pre_event_names}')
        await asyncio.wait([asyncio.create_task(event.call()) for event in pre_events], timeout=self.stage_timeout)

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Starting Teardown stage.')

        methods = inspect.getmembers(self, predicate=inspect.ismethod) 
        for _, method in methods:

            method_name = method.__qualname__
            hook: Hook = registrar.all.get(method_name)

            if hook:
                self.hooks[hook.hook_type].append(hook)

        teardown_hooks: List[TeardownHook] = self.hooks[HookType.TEARDOWN]

        if teardown_hooks:

            teardown_hook_names = ', '.join([hook.name for hook in teardown_hooks])

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Running teardown hooks - {teardown_hook_names}')           
            await asyncio.gather(*[ hook.call() for hook in teardown_hooks])

        post_events: List[EventHook] = [
            event for event in events if isinstance(event, EventHook) and event.pre is False
        ]

        if len(post_events) > 0:
            post_event_names = ", ".join([
                event.shortname for event in post_events
            ])

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing POST events - {post_event_names}')
            await asyncio.wait([
                asyncio.create_task(event.call()) for event in post_events
            ], timeout=self.stage_timeout)

        context_hooks: List[ContextHook] = self.hooks[HookType.CONTEXT]
        context_hooks: List[ContextHook] = self.hooks[HookType.CONTEXT]
        await asyncio.gather(*[
            asyncio.create_task(context_hook.call(self.context)) for context_hook in context_hooks
        ])

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Teardown complete.')



from .headers_sent_event import HeadersSent


class TrailersSent(HeadersSent):
    event_type='TRAILERS_SENT'
    """
    The _TrailersSent event is fired whenever trailers are sent on a
    stream. Trailers are a set of headers sent after the body of the
    request/response, and are used to provide information that wasn't known
    ahead of time (e.g. content-length).

    This is an internal event, used to determine validation steps on
    outgoing header blocks.
    """
    pass



from typing import Any, Dict, Type, Callable, Awaitable, Any
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from .hook import Hook
from .hook_metadata import HookMetadata


class TeardownHook(Hook):

    def __init__(
        self, 
        name: str, 
        shortname: str, 
        call: Callable[..., Awaitable[Any]], 
        metadata: Dict[str, Any]={}
    ) -> None:

        if metadata is None:
            metadata = {}

        super().__init__(
            name, 
            shortname, 
            call, 
            hook_type=HookType.TEARDOWN
        )

        self.call: Type[self._call] = self._call
        self.metadata = HookMetadata(**metadata)


import asyncio
import functools
import os
import psutil
import uuid
from typing import List
from numpy import float32, float64, int16, int32, int64
from concurrent.futures import ThreadPoolExecutor
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet


try:
    import uuid
    from cassandra.cqlengine import columns
    from cassandra.cqlengine import connection
    from datetime import datetime
    from cassandra.cqlengine.management import sync_table
    from cassandra.cqlengine.models import Model
    from cassandra.cluster import Cluster
    from cassandra.auth import PlainTextAuthProvider
    from .cassandra_config import CassandraConfig
    has_connector = True

except Exception:
    Cluster = None
    PlainTextAuthProvider = None
    CassandraConfig = None
    has_connector = False


class Cassandra:

    def __init__(self, config: CassandraConfig) -> None:
        self.cluster = None
        self.session = None

        self.hosts = config.hosts
        self.port = config.port or 9042
        self.username = config.username
        self.password = config.password
        self.keyspace = config.keyspace
        self.custom_fields = config.custom_fields
        self.events_table_name: str = config.events_table
        self.metrics_table_name: str = config.metrics_table
        self.shared_metrics_table_name = 'stage_metrics'
        self.custom_metrics_table_names = {}
        self.errors_table_name = 'stage_errors'
        self.replication_strategy = config.replication_strategy
        self.replication = config.replication       
        self.ssl = config.ssl

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self._metrics_table = None
        self._errors_table = None
        self._events_table = None
        self._shared_metrics_table = None
        self._custom_metrics_tables = {}

        self._executor =ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        self._loop = asyncio.get_event_loop()

    async def connect(self):

        host_port_combinations = ', '.join([
            f'{host}:{self.port}' for host in self.hosts
        ])

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opening amd authorizing connection to Cassandra Cluster at - {host_port_combinations}')
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Opening session - {self.session_uuid}')

        auth = None
        if self.username and self.password:
            auth = PlainTextAuthProvider(self.username, self.password)
        

        self.cluster = Cluster(
            self.hosts,
            port=self.port,
            auth_provider=auth,
            ssl_context=self.ssl
        )

        self.session = await self._loop.run_in_executor(
            None,
            self.cluster.connect
        )


        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to Cassandra Cluster at - {host_port_combinations}')

        if self.keyspace is None:
            self.keyspace = 'hedra'

        await self.logger.filesystem.aio['hedra.repoorting'].info(f'{self.metadata_string} - Creating Keyspace - {self.keyspace}')

        keyspace_options = f"'class' : '{self.replication_strategy}', 'replication_factor' : {self.replication}"
        keyspace_query = f"CREATE KEYSPACE IF NOT EXISTS {self.keyspace} WITH REPLICATION = " + "{" + keyspace_options  + "};"

        await self._loop.run_in_executor(
            None,
            self.session.execute,
            keyspace_query
        )

        await self._loop.run_in_executor(
            None,
            self.session.set_keyspace,
            self.keyspace
        )
        if os.getenv('CQLENG_ALLOW_SCHEMA_MANAGEMENT') is None:
            os.environ['CQLENG_ALLOW_SCHEMA_MANAGEMENT'] = '1'

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                connection.setup,
                self.hosts, 
                self.keyspace, 
                protocol_version=3
            )
        )

        await self.logger.filesystem.aio['hedra.repoorting'].info(f'{self.metadata_string} - Created Keyspace - {self.keyspace}')

    async def submit_events(self, events: List[BaseEvent]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to - Keyspace: {self.keyspace} - Table: {self.events_table_name}')

        if self._events_table is None:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Events table - {self.events_table_name} - under keyspace - {self.keyspace}')

            self._events_table = type(
                self.events_table_name.capitalize(), 
                (Model, ), 
                {
                    'id': columns.UUID(primary_key=True, default=uuid.uuid4),
                    'name': columns.Text(min_length=1, index=True),
                    'stage': columns.Text(min_length=1),
                    'time': columns.Float(),
                    'succeeded': columns.Boolean(),
                    'created_at': columns.DateTime(default=datetime.now)
                }
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    sync_table,
                    self._events_table,
                    keyspaces=[self.keyspace]
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Events table - {self.events_table_name} - under keyspace - {self.keyspace}')

        for event in events:
            
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self._metrics_table.create,
                    **event.record
                )
            )

        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to - Keyspace: {self.keyspace} - Table: {self.events_table_name}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to - Keyspace: {self.keyspace} - Table: {self.shared_metrics_table_name}')
        
        if self._shared_metrics_table is None:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Shared Metrics table - {self.shared_metrics_table_name} - under keyspace - {self.keyspace}')

            fields = {
                'id': columns.UUID(primary_key=True, default=uuid.uuid4),
                'name': columns.Text(min_length=1, index=True),
                'stage': columns.Text(min_length=1),
                'group': columns.Text(min_length=1),
                'total': columns.Integer(),
                'succeeded': columns.Integer(),
                'failed': columns.Integer(),
                'actions_per_second': columns.Float()
            }

            self._shared_metrics_table = type(
                self.shared_metrics_table_name.capitalize(), 
                (Model, ), 
                fields
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    sync_table,
                    self._shared_metrics_table,
                    keyspaces=[self.keyspace]
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Shared Metrics table - {self.shared_metrics_table_name} - under keyspace - {self.keyspace}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self._metrics_table.create,
                        **{
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            'group': 'common',
                            **metrics_set.common_stats
                        }
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to - Keyspace: {self.keyspace} - Table: {self.shared_metrics_table_name}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to - Keyspace: {self.keyspace} - Table: {self.metrics_table_name}')

        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
       
            if self._metrics_table is None:
                
                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Metrics table - {self.metrics_table_name} - under keyspace - {self.keyspace}')
                
                fields = {
                    'id': columns.UUID(primary_key=True, default=uuid.uuid4),
                    'name': columns.Text(min_length=1, index=True),
                    'stage': columns.Text(min_length=1),
                    'group': columns.Text(),
                    'median': columns.Float(),
                    'mean': columns.Float(),
                    'variance': columns.Float(),
                    'stdev': columns.Float(),
                    'minimum': columns.Float(),
                    'maximum': columns.Float(),
                    'created_at': columns.DateTime(default=datetime.now)
                }


                for quantile_name in metrics_set.quantiles:
                    fields[quantile_name] = columns.Float()
                
                for custom_field_name, custom_field_type in metrics_set.custom_schemas:
                    fields[custom_field_name] = custom_field_type


                self._metrics_table = type(
                    self.metrics_table_name.capitalize(), 
                    (Model, ), 
                    fields
                )

                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        sync_table,
                        self._metrics_table,
                        keyspaces=[self.keyspace]
                    )
                )

                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Metrics table - {self.metrics_table_name} - under keyspace - {self.keyspace}')

            for group_name, group in metrics_set.groups.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Group - {group_name}:{group.metrics_group_id}')                
                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self._metrics_table.create,
                        **{
                            **group.record,
                            'group': group_name
                        }
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics Set to - Keyspace: {self.keyspace} - Table: {self.metrics_table_name}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics Set to - Keyspace: {self.keyspace} - Table: {self.metrics_table_name}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            
            for custom_group_name, group in metrics_set.custom_metrics.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Group - {custom_group_name}')

                custom_metrics_table_name = f'{custom_group_name}_metrics'
                if self._custom_metrics_tables.get(custom_metrics_table_name):

                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Custom Metrics table - {custom_metrics_table_name} - under keyspace - {self.keyspace}')
                    
                    custom_table = {
                        'id': columns.UUID(primary_key=True, default=uuid.uuid4),
                        'name': columns.Text(min_length=1, index=True),
                        'stage': columns.Text(min_length=1),
                        'group': columns.Text(min_length=1)
                    }

                    for field_name, value in group.items():

                        cassandra_column = None
                        if isinstance(value, (int, int16, int32, int64)):
                            cassandra_column = columns.Integer()
                        
                        elif isinstance(value, (float, float32, float64)):
                            cassandra_column = columns.Float()

                        custom_table[field_name] = cassandra_column

                    custom_metrics_table = type(
                        custom_metrics_table_name.capitalize(), 
                        (Model, ), 
                        custom_table
                    )

                    await self._loop.run_in_executor(
                        self._executor,
                        functools.partial(
                            sync_table,
                            custom_metrics_table,
                            keyspaces=[self.keyspace]
                        )
                    )

                    self._custom_metrics_tables[custom_metrics_table_name] = custom_metrics_table

                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Custom Metrics table - {custom_metrics_table_name} - under keyspace - {self.keyspace}')
    
                table = self._custom_metrics_tables.get(custom_metrics_table_name)
                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        table.create,
                        {
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            'group': custom_group_name,
                            **group
                        }
                    )
                )
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics Set to - Keyspace: {self.keyspace} - Table: {self.metrics_table_name}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Errors Metrics to - Keyspace: {self.keyspace} - Table: {self.errors_table_name}')

        if self._errors_table is None:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Errors Metrics table - {self.errors_table_name} - under keyspace - {self.keyspace}')

            errors_table_fields = {
                'id': columns.UUID(primary_key=True, default=uuid.uuid4),
                'name': columns.Text(min_length=1, index=True),
                'stage': columns.Text(min_length=1),
                'error_message': columns.Text(min_length=1),
                'error_count': columns.Integer()
            }

            self._errors_table = type(
                self.errors_table_name.capitalize()
                (Model, ),
                errors_table_fields
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    sync_table,
                    self._errors_table,
                    keyspaces=[self.keyspace]
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Errors Metrics table - {self.errors_table_name} - under keyspace - {self.keyspace}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Errors Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self._errors_table.create,
                        {
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            'error_message': error.get('message'),
                            'error_count': error.get('count')
                        }
                    )
                )
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Errors Metrics to - Keyspace: {self.keyspace} - Table: {self.errors_table_name}')
 
    async def close(self):
        host_port_combinations = ', '.join([
            f'{host}:{self.port}' for host in self.hosts
        ])

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closing connection to Cassandra Cluster at - {host_port_combinations}')

        await self._loop.run_in_executor(
            None,
            self.cluster.shutdown
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closed connection to Cassandra Cluster at - {host_port_combinations}')


from random import randrange
from typing import List
from hedra.core.engines.types.common.timeouts import Timeouts
from hedra.core.engines.types.common.types import RequestTypes
from .pipe import HTTP2Pipe
from .connection import HTTP2Connection


class HTTP2Pool:

    __slots__ = (
        'size',
        'connections',
        'pipes',
        'timeouts',
        'reset_connections',
        'pool_type'
    )

    def __init__(self, size: int, timeouts: Timeouts, reset_connections: bool=False) -> None:
        self.size = size
        self.connections: List[HTTP2Connection] = []
        self.pipes: List[HTTP2Pipe] = []
        self.timeouts = timeouts
        self.reset_connections = reset_connections
        self.pool_type: RequestTypes = RequestTypes.HTTP2

    def create_pool(self) -> None:

        self.pipes = [ HTTP2Pipe(self.size) for _ in range(self.size) ]
        
        self.connections = [
            HTTP2Connection(
                randrange(1, 2**20 + 2, 2), 
                self.timeouts, 
                self.size, 
                self.reset_connections,
                self.pool_type
            ) for _ in range(0, self.size * 2, 2)
        ]

    def reset(self):
        self.pipes.append(HTTP2Pipe(self.size))

        self.connections.append(
            HTTP2Connection(
                randrange(1, 2**20 + 2, 2), 
                self.timeouts, 
                self.size, 
                self.reset_connections,
                self.pool_type
            )
        )

    async def close(self):
        for connection in self.connections:
            await connection.close()



import asyncio
import traceback
import h2.stream
import h2.config
import h2.connection
import h2.events
import h2.exceptions
import h2.settings
import h2.errors
from hedra.core.engines.types.http2.stream import Stream
from hedra.core.engines.types.common.encoder import Encoder
from hedra.core.engines.types.common.decoder import Decoder
from hedra.core.engines.types.http2.events.remote_settings_changed_event import RemoteSettingsChanged
from hedra.core.engines.types.http2.events.settings_acknowledged_event import SettingsAcknowledged
from hedra.core.engines.types.http2.streams.stream_settings import SettingCodes
from hedra.core.engines.types.http2.events.stream_reset import StreamReset
from hedra.core.engines.types.http2.streams.stream_closed_by import StreamClosedBy
from hedra.core.engines.types.http2.events.deferred_headers_event import DeferredHeaders
from hedra.core.engines.types.http2.events.connection_terminated_event import ConnectionTerminated
from hedra.core.engines.types.http2.errors.exceptions import StreamClosedError
from hedra.core.engines.types.http2.events.data_received_event import DataReceived
from hedra.core.engines.types.http2.errors.exceptions import StreamError
from hedra.core.engines.types.http2.errors.types import ErrorCodes
from hedra.core.engines.types.http2.events.window_updated_event import WindowUpdated
from .action import HTTP2Action
from .result import HTTP2Result
from .windows import WindowManager
from .frames.types.base_frame import Frame


class HTTP2Pipe:
    __slots__ = (
        '_h2_state',
        'connected',
        'concurrency',
        '_encoder',
        '_decoder',
        '_init_sent',
        'stream_id',
        '_data_to_send',
        '_headers_sent',
        'lock',
        'local_settings',
        'remote_settings',
        'outbound_flow_control_window',
        'local_settings',
        '_inbound_flow_control_window_manager',
        'local_settings_dict',
        'remote_settings_dict'
    )

    CONFIG = h2.config.H2Configuration(
        validate_inbound_headers=False,
    )

    def __init__(self, concurrency):
        self._h2_state = h2.connection.H2Connection(config=self.CONFIG)
        self.connected = False
        self.concurrency = concurrency
        self._encoder = Encoder()
        self._decoder = Decoder()
        self._init_sent = False
        self.stream_id = None
        self._data_to_send = b''
        self._headers_sent = False
        self.lock = asyncio.Lock()

        self.local_settings = h2.settings.Settings(
            client=True,
            initial_values={
                h2.settings.SettingCodes.ENABLE_PUSH: 0,
                h2.settings.SettingCodes.MAX_CONCURRENT_STREAMS: concurrency,
                h2.settings.SettingCodes.MAX_HEADER_LIST_SIZE: 65535,
            }
        )
        self.remote_settings = h2.settings.Settings(
            client=False
        )

        self.outbound_flow_control_window = self.remote_settings.initial_window_size

        del self.local_settings[h2.settings.SettingCodes.ENABLE_CONNECT_PROTOCOL]

        self._inbound_flow_control_window_manager = WindowManager(
            max_window_size=self.local_settings.initial_window_size
        )

        self.local_settings_dict = {setting_name: setting_value for setting_name, setting_value in self.local_settings.items()}
        self.remote_settings_dict = {setting_name: setting_value for setting_name, setting_value in self.remote_settings.items()}


    def _guard_increment_window(self, current, increment):
        # The largest value the flow control window may take.
        LARGEST_FLOW_CONTROL_WINDOW = 2**31 - 1

        new_size = current + increment

        if new_size > LARGEST_FLOW_CONTROL_WINDOW:
            self.outbound_flow_control_window = (
                self.remote_settings.initial_window_size
            )

            self._inbound_flow_control_window_manager = WindowManager(
                max_window_size=self.local_settings.initial_window_size
            )

        return LARGEST_FLOW_CONTROL_WINDOW - current

    def send_request_headers(self, request: HTTP2Action, stream: Stream):

        self._headers_sent = False

        if self._init_sent is False or stream.reset_connection:

            window_increment = 65536

            self._inbound_flow_control_window_manager.window_opened(window_increment)

            stream.write(bytes(stream.connection_data))
            self._init_sent = True

            self.outbound_flow_control_window = self.remote_settings.initial_window_size

        stream.inbound = WindowManager(self.local_settings.initial_window_size)
        stream.outbound = WindowManager(self.remote_settings.initial_window_size)
        stream.max_inbound_frame_size = self.local_settings.max_frame_size
        stream.max_outbound_frame_size = self.remote_settings.max_frame_size
        stream.current_outbound_window_size = self.remote_settings.initial_window_size

        end_stream = request.encoded_data is None
        encoded_headers = request.encoded_headers

        if request.hpack_encoder.header_table.maxsize != self._encoder.header_table.maxsize:
            request.hpack_encoder.header_table_size = self._encoder.header_table.maxsize
            request._setup_headers()
    
        stream.headers_frame.data = encoded_headers[0]
        headers_frame = stream.headers_frame
        if end_stream:
            headers_frame.flags.add('END_STREAM')

        stream.inbound.window_opened(65536) 
  
        stream.write(headers_frame.serialize())
        self._headers_sent = True

    async def receive_response(self, response: HTTP2Result, stream: Stream):

        done = False
        while done is False:
            
            data = await asyncio.wait_for(stream.read(), timeout=stream.timeouts.socket_read_timeout)

            stream.frame_buffer.data.extend(data)
            stream.frame_buffer.max_frame_size = stream.max_outbound_frame_size

            write_data = bytearray()
            frames = None
            stream_events = []

            for frame in stream.frame_buffer:
                try:

                    if frame.type == 0x0:
                        # DATA

                        end_stream = 'END_STREAM' in frame.flags
                        flow_controlled_length = frame.flow_controlled_length
                        frame_data = frame.data

                        frames = []
                        self._inbound_flow_control_window_manager.window_consumed(
                            flow_controlled_length
                        )

                        try:
                            
                            stream.inbound.window_consumed(flow_controlled_length)
                    
                            event = DataReceived()
                            event.stream_id = stream.stream_id

                            stream_events.append(event)

                            if end_stream:
                                done = True

                            stream_events[0].data = frame_data
                            stream_events[0].flow_controlled_length = flow_controlled_length

                        except StreamClosedError as e:
                            raise Exception(f'Connection - {stream.stream_id} err: {str(e._events[0])}')

                    elif frame.type == 0x07:
                        # GOAWAY
                        self._data_to_send = b''

                        new_event = ConnectionTerminated()
                        new_event.error_code = h2.errors._error_code_from_int(frame.error_code)
                        new_event.last_stream_id = frame.last_stream_id
                        
                        if frame.additional_data:
                            new_event.additional_data = frame.additional_data

                        frames = []
                        raise Exception(f'Connection - {stream.stream_id} err: {str(new_event)}')

                    elif frame.type == 0x01:
                        # HEADERS
                        deferred_headers = DeferredHeaders(
                            stream.encoder,
                            frame,
                            self._h2_state.config.header_encoding
                        )

                        if deferred_headers.end_stream:
                            done = True

                        frames = []
                        response.deferred_headers = deferred_headers

                    elif frame.type == 0x03:
                        # RESET

                        stream.closed_by = StreamClosedBy.RECV_RST_STREAM
                        reset_event = StreamReset()
                        reset_event.stream_id = stream.stream_id

                        reset_event.error_code = h2.errors._error_code_from_int(frame.error_code)

                        raise Exception(f'Connection - {stream.stream_id} err: {str(reset_event)}')

                    elif frame.type == 0x04:
                        # SETTINGS

                        if 'ACK' in frame.flags:

                            changes = self.local_settings.acknowledge()
                    
                            initial_window_size_change = changes.get(SettingCodes.INITIAL_WINDOW_SIZE)
                            max_header_list_size_change = changes.get(SettingCodes.MAX_HEADER_LIST_SIZE)
                            max_frame_size_change = changes.get(SettingCodes.MAX_FRAME_SIZE)
                            header_table_size_change =changes.get(SettingCodes.HEADER_TABLE_SIZE)

                            if initial_window_size_change is not None:

                                window_delta = initial_window_size_change.new_value - initial_window_size_change.original_value
                                
                                new_max_window_size = stream.inbound.max_window_size + window_delta
                                stream.inbound.window_opened(window_delta)
                                stream.inbound.max_window_size = new_max_window_size

                            if max_header_list_size_change is not None:
                                self._decoder.max_header_list_size = max_header_list_size_change.new_value

                            if max_frame_size_change is not None:
                                stream.max_outbound_frame_size =  max_frame_size_change.new_value

                            if header_table_size_change:
                                # This is safe across all hpack versions: some versions just won't
                                # respect it.
                                self._decoder.max_allowed_table_size = header_table_size_change.new_value

                        # Add the new settings.
                        self.remote_settings.update(frame.settings)
           
                        changes = self.remote_settings.acknowledge()
                        initial_window_size_change = changes.get(SettingCodes.INITIAL_WINDOW_SIZE)
                        header_table_size_change = changes.get(SettingCodes.HEADER_TABLE_SIZE)
                        max_frame_size_change = changes.get(SettingCodes.MAX_FRAME_SIZE)
                    
                        if initial_window_size_change:
                            stream.current_outbound_window_size = self._guard_increment_window(
                                stream.current_outbound_window_size,
                                initial_window_size_change.new_value - initial_window_size_change.original_value
                            )

                        # HEADER_TABLE_SIZE changes by the remote part affect our encoder: cf.
                        # RFC 7540 Section 6.5.2.
                        if  header_table_size_change:
                            self._encoder.header_table_size = header_table_size_change.new_value

                        if max_frame_size_change:
                            stream.max_outbound_frame_size = max_frame_size_change.new_value

                        frame = Frame(0, 0x04)
                        frame.flags.add('ACK')

                        frames = [frame]

                    elif frame.type == 0x08:
                        # WINDOW UPDATE

                        frames = []
                        increment = frame.window_increment
                        if frame.stream_id:
                            try:

                                
                                event = WindowUpdated()
                                event.stream_id = stream.stream_id

                                # If we encounter a problem with incrementing the flow control window,
                                # this should be treated as a *stream* error, not a *connection* error.
                                # That means we need to catch the error and forcibly close the stream.
                                event.delta = increment

                                try:
                                    self.outbound_flow_control_window = self._guard_increment_window(
                                        self.outbound_flow_control_window,
                                        increment
                                    )
                                except StreamError:
                                    # Ok, this is bad. We're going to need to perform a local
                                    # reset.

                                    event = StreamReset()
                                    event.stream_id = stream.stream_id
                                    event.error_code = ErrorCodes.FLOW_CONTROL_ERROR
                                    event.remote_reset = False

                                    stream.closed_by = ErrorCodes.FLOW_CONTROL_ERROR    
                      
                                    raise Exception(f'Connection - {stream.stream_id} err: {str(event)}')
                            except Exception:
                                frames = []
                        else:
                            self.outbound_flow_control_window = self._guard_increment_window(
                                self.outbound_flow_control_window,
                                increment
                            )
                            # FIXME: Should we split this into one event per active stream?
                            window_updated_event = WindowUpdated()
                            window_updated_event.stream_id = 0
                            window_updated_event.delta = increment

                            frames = []

                except Exception as e:
                    raise Exception(f'Connection {stream.stream_id} err- {str(e)}')

                if frames:
                    for f in frames:
                        write_data.extend(f.serialize())

                    stream.write(write_data)

            for event in stream_events:
                amount = event.flow_controlled_length

                conn_increment = self._inbound_flow_control_window_manager.process_bytes(amount)

                if conn_increment:
                    stream.write_window_update_frame(0, conn_increment)

                response.body.extend(event.data)

            if done:
                break

        return response

    async def submit_request_body(self, request: HTTP2Action, stream: Stream) -> None:
        data = request.encoded_data
        
        while data:
            local_flow = stream.current_outbound_window_size
            max_frame_size = stream.max_outbound_frame_size
            flow = min(local_flow, max_frame_size)
            while flow == 0:
                await self._receive_events(stream)
                local_flow = stream.current_outbound_window_size
                max_frame_size = stream.max_outbound_frame_size
                flow = min(local_flow, max_frame_size)
                
            max_flow = flow
            chunk_size = min(len(data), max_flow)
            chunk, data = data[:chunk_size], data[chunk_size:]

            df = Frame(stream.stream_id, 0x0)
            df.data = chunk

            # Subtract flow_controlled_length to account for possible padding
            self.outbound_flow_control_window -= df.flow_controlled_length
            assert self.outbound_flow_control_window >= 0

            stream.write(df.serialize())

        df = Frame(stream.stream_id, 0x0)
        df.flags.add('END_STREAM')

        stream.write(df.serialize())


from typing import Union
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.common.base_result import BaseResult
from .action import UDPAction


class UDPResult(BaseResult):

    __slots__ = (
        'action_id',
        'url',
        'ip_addr',
        'has_response',
        'path',
        'params',
        'query',
        'hostname',
        'body',
        'response_code',
        '_version',
        '_reason'
        '_status'
    )

    def __init__(self, action: UDPAction, error: Exception=None) -> None:

        super(
            UDPResult,
            self
        ).__init__(
            action.action_id,
            action.name,
            action.url.hostname,
            action.metadata.user,
            action.metadata.tags,
            RequestTypes.UDP,
            action.hooks.checks,
            error
        )

        self.url = action.url.full
        self.ip_addr = action.url.ip_addr
        self.has_response = action.wait_for_response
        self.path = action.url.path
        self.params = action.url.params
        self.query = action.url.query
        self.hostname = action.url.hostname

        self.body = bytearray()
        self.response_code = None
        self._version = None
        self._reason = None
        self._status = None

    @property
    def size(self):
        
        if self.body:
            return len(self.body)
        
        else:
            return 0

    @property
    def data(self) -> Union[str, dict, None]:
        return self.body.decode()

    @property
    def status(self) -> Union[int, None]:
        return self._status

    @status.setter
    def status(self, new_status: int):
        self._status = new_status

    def to_dict(self):

        base_result_dict = super().to_dict()

        data = self.data
        if isinstance(data, bytes) or isinstance(data, bytearray):
            data = str(data.decode())
        
        return {
            'url': self.url,
            'path': self.path,
            'params': self.params,
            'query': self.query,
            'type': self.type,
            'data': data,
            'tags': self.tags,
            'user': self.user,
            'error': str(self.error),
            'status': self.status,
            **base_result_dict
        }


import os
from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    SQLiteConfig
)


class SubmitSQLiteResultsStage(Submit):
    config=SQLiteConfig(
        path=f'{os.getcwd}/results.db',
        events_table='events',
        metrics_table='metrics'
    )


from hedra.core.graphs.stages import (
    Execute,
)

from hedra.core.graphs.hooks import (
    action
)


class ExecuteWebsocketStage(Execute):

    @action()
    async def webbsocket_send(self):
        return await self.client.websocket.send('https://<url_here>', data={"PING": "PONG"})

    @action()
    async def webbsocket_listen(self):
        return await self.client.websocket.listen('https://<url_here>')


from typing import Union
from hedra.core.graphs.hooks.registry.registry_types import (
    ActionHook,
    TaskHook
)



class CompletedCounter:

    def __init__(self) -> None:
        self.completed_count = 0
        self.last_completed = 0
        self.last_batch_size = 0
        self.action = None

    async def execute_action(self, hook: Union[ActionHook, TaskHook]):
        response = await hook.session.execute_prepared_request(
            hook.action
        )
        
        
        self.completed_count += 1

        return response


from __future__ import annotations
import os
import sys
import asyncio
import aiologger
from asyncio import Task
from aiologger.levels import LogLevel
from aiologger.formatters.base import Formatter
from aiologger.handlers.streams import AsyncStreamHandler
from .logger_types import LoggerTypes


class AsyncLogger(aiologger.Logger):

    def __init__(
        self, 
        logger_name: str=None, 
        logger_type: LoggerTypes=LoggerTypes.CONSOLE,
        log_level: LogLevel = LogLevel.INFO, 
        logger_enabled: bool=True
    ) -> None:
        super().__init__(name=logger_name, level=log_level)

        self.logger_name = logger_name
        self.logger_type = logger_type
        self.log_level = log_level
        self.logger_enabled = logger_enabled
        self.pattern = None
        self.datefmt_pattern = None

    def initialize(self, pattern: str, datefmt_pattern: str=None):

        self.pattern = pattern
        self.datefmt_pattern = datefmt_pattern

        self.add_handler(
            AsyncStreamHandler(
                stream=os.fdopen(
                    os.dup(sys.__stdout__.fileno())
                ),
                level=self.level,
                formatter=Formatter(pattern, datefmt=datefmt_pattern),
            )
        )

    def _make_log_task(self, level, msg, *args, **kwargs) -> Task:

        if self.logger_enabled:
            return super()._make_log_task(level, msg, *args, **kwargs)

        else:
            return asyncio.create_task(self._skip_task())

    async def _skip_task(self):
        return





import os
from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    CosmosDBConfig
)


class SubmitCosmosDBResultsStage(Submit):
    config=CosmosDBConfig(
        account_uri=os.getenv('AZURE_COSMOSDB_ACCOUNT_URI', ''),
        account_key=os.getenv('AZURE_COSMOSDB_ACCOUNT_KEY', ''),
        database='results',
        events_container='events',
        metrics_container='metrics',
        events_partition_key='name',
        metrics_partition_key='name'
    )


from typing import Callable, Awaitable, Any
from hedra.core.graphs.simple_context import SimpleContext
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.tools.filesystem import open
from .hook import Hook


class RestoreHook(Hook):

    def __init__(
        self, 
        name: str, 
        shortname: str, 
        call: Callable[..., Awaitable[Any]], 
        key: str=None,
        restore_filepath: str=None
    ) -> None:
        super().__init__(
            name, 
            shortname, 
            call, 
            hook_type=HookType.RESTORE
        )

        self.context_key = key
        self.restore_path = restore_filepath

    async def call(self, context: SimpleContext) -> None:

        restore_file = await open(self.restore_path, 'r')
        context[self.context_key] = await self._call(
            await restore_file.read()
        )

        await restore_file.close()



import functools
from .hook_type import HookType
from hedra.core.graphs.hooks.registry.registrar import registrar


@registrar(HookType.VALIDATE)
def validate(stage: str, *names):
    
    def wrapper(func):

        @functools.wraps(func)
        def decorator(*args, **kwargs):
            return func(*args, **kwargs)

        return decorator

    return wrapper


import os
from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    DatadogConfig
)


class SubmitDatadogResultsStage(Submit):
    config=DatadogConfig(
        api_key=os.getenv('DATADOG_API_KEY', ''),
        app_key=os.getenv('DATADOG_APP_KEY', '')
    )


try:
    from hedra.reporting.types.statsd.statsd import StatsD
    from .netdata_config import NetdataConfig
    has_connector = True

except Exception:
    from hedra.reporting.types.empty import Empty as StatsD
    NetdataConfig = None
    has_connector = False


class Netdata(StatsD):

    def __init__(self, config: NetdataConfig) -> None:
        super(Netdata, self).__init__(config)
        self.statsd_type = 'Netdata'


import asyncio
import time
import uuid
import traceback
from typing import Awaitable, Dict, Set, Tuple, Union
from hedra.core.engines.types.common.ssl import get_default_ssl_context
from hedra.core.engines.types.common.timeouts import Timeouts
from hedra.core.engines.types.common.concurrency import Semaphore
from hedra.logging import HedraLogger
from .connection import HTTPConnection
from .action import HTTPAction
from .result import HTTPResult
from .pool import Pool


HTTPResponseFuture = Awaitable[Union[HTTPResult, Exception]]
HTTPBatchResponseFuture = Awaitable[Tuple[Set[HTTPResponseFuture], Set[HTTPResponseFuture]]]


class MercuryHTTPClient:

    __slots__ = (
        'session_id',
        'timeouts',
        'registered',
        '_hosts',
        'closed',
        'sem',
        'pool',
        'active',
        'waiter',
        'ssl_context',
        'logger'
    )

    def __init__(self, concurrency: int=10**3, timeouts: Timeouts = Timeouts(), reset_connections: bool=False) -> None:

        self.session_id = str(uuid.uuid4())
        self.timeouts = timeouts

        self.registered: Dict[str, HTTPAction] = {}
        self._hosts = {}
        self.closed = False

        self.sem = asyncio.Semaphore(value=concurrency)
        self.pool = Pool(concurrency, reset_connections=reset_connections)
        self.pool.create_pool()
        self.active = 0
        self.waiter = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self.ssl_context = get_default_ssl_context()

    async def wait_for_active_threshold(self):
        if self.waiter is None:
            self.waiter = asyncio.get_event_loop().create_future()
            await self.waiter


    async def prepare(self, action: HTTPAction) -> Awaitable[Union[HTTPAction, Exception]]:
        try:
            if action.url.is_ssl:
                action.ssl_context = self.ssl_context
                
            if self._hosts.get(action.url.hostname) is None:
                socket_configs = await asyncio.wait_for(action.url.lookup(), timeout=self.timeouts.connect_timeout)
            
                for ip_addr, configs in socket_configs.items():
                    for config in configs:

                        connection = HTTPConnection()
                        
                        try:
                            await connection.make_connection(
                                action.url.hostname,
                                ip_addr,
                                action.url.port,
                                config,
                                ssl=action.ssl_context,
                                timeout=self.timeouts.connect_timeout
                            )

                            action.url.socket_config = config
                            action.url.ip_addr = ip_addr
                            action.url.has_ip_addr = True
                            break

                        except Exception as e:
                            pass

                    if action.url.socket_config:
                        break

                if action.url.socket_config is None:
                    raise Exception('Err. - No socket found.')

                self._hosts[action.url.hostname] = {
                    'ip_addr': action.url.ip_addr,
                    'socket_config': action.url.socket_config
                }

            else:
                host_config = self._hosts[action.url.hostname]
                action.url.ip_addr = host_config.get('ip_addr')
                action.url.socket_config = host_config.get('socket_config')

            if action.is_setup is False:
                action.setup()

            self.registered[action.name] = action

        except Exception as e:       
            raise e

    def extend_pool(self, increased_capacity: int):
        self.pool.size += increased_capacity
        for _ in range(increased_capacity):
            self.pool.connections.append(
                HTTPConnection(self.pool.reset_connections)
            )
        
        self.sem = Semaphore(self.pool.size)

    def shrink_pool(self, decrease_capacity: int):
        self.pool.size -= decrease_capacity
        self.pool.connections = self.pool.connections[:self.pool.size]
        self.sem = Semaphore(self.pool.size)

    async def execute_prepared_request(self, action: HTTPAction) -> HTTPResponseFuture:
  
        response = HTTPResult(action)
        response.wait_start = time.monotonic()
        self.active += 1
 
        async with self.sem:
            connection = self.pool.connections.pop()
            
            try:
                
                if action.hooks.listen:
                    event = asyncio.Event()
                    action.hooks.channel_events.append(event)
                    await event.wait()

                if action.hooks.before:
                    action = await action.hooks.before(action, response)
                    action.setup()

                response.start = time.monotonic()

                await connection.make_connection(
                    action.url.hostname,
                    action.url.ip_addr,
                    action.url.port,
                    action.url.socket_config,
                    timeout=self.timeouts.connect_timeout,
                    ssl=action.ssl_context
                )

                response.connect_end = time.monotonic()

                connection.write(action.encoded_headers)
                
                if action.encoded_data:
                    if action.is_stream:
                        action.write_chunks(connection)

                    else:
                        connection.write(action.encoded_data)

                response.write_end = time.monotonic()

                response.response_code = await connection.reader.readline_fast()
                headers = await connection.read_headers()
                status = response.status

                if status >= 300 and status < 400:
                    for _ in range(action.redirects):

                        redirect_url = headers.get(b'Location')

                        redirect_action = HTTPAction(
                            action.name,
                            redirect_url,
                            method=action.method
                        )

                        redirect_action.is_setup = True

                        self.prepare(redirect_action)

                        redirect_action.encoded_headers = action.encoded_headers
                        redirect_action.encoded_data = action.encoded_data

                        await connection.make_connection(
                            redirect_action.url.hostname,
                            redirect_action.url.ip_addr,
                            redirect_action.url.port,
                            redirect_action.url.socket_config,
                            timeout=self.timeouts.connect_timeout,
                            ssl=redirect_action.ssl_context
                        )

                        response.connect_end = time.monotonic()

                        connection.write(redirect_action.encoded_headers)
                        
                        if redirect_action.encoded_data:
                            if redirect_action.is_stream:
                                redirect_action.write_chunks(connection)

                            else:
                                connection.write(redirect_action.encoded_data)

                        response.write_end = time.monotonic()

                        response.response_code = await connection.reader.readline_fast()
                        headers = await connection.read_headers()

                        status = response.status

                        if status >= 200 and status < 300:
                            break

                content_length = headers.get(b'content-length')
                transfer_encoding = headers.get(b'transfer-encoding')

                # We require Content-Length or Transfer-Encoding headers to read a
                # request body, otherwise it's anyone's guess as to how big the body
                # is, and we ain't playing that game.
                body = bytearray()
                if content_length:
                    body = await connection.readexactly(int(content_length))

                elif transfer_encoding:
                    
                    all_chunks_read = False

                    while True and not all_chunks_read:

                        chunk_size = int((await connection.readuntil()).rstrip(), 16)
    
                        if not chunk_size:
                            # read last CRLF
                            body.extend(
                                await connection.readuntil()
                            )
                            break
                        
                        chunk = await connection.readexactly(chunk_size + 2)
                        body.extend(
                            chunk[:-2]
                        )

                    all_chunks_read = True
         
                response.complete = time.monotonic()
                response.headers = headers
                response.body = body
                
                self.pool.connections.append(connection)

                if action.hooks.after:
                    action = await action.hooks.after(action, response)
                    action.setup()

                if action.hooks.notify:
                    await asyncio.gather(*[
                        asyncio.create_task(
                            channel(response, action.hooks.listeners)
                        ) for channel in action.hooks.channels
                    ])

                    for listener in action.hooks.listeners: 
                        if len(listener.hooks.channel_events) > 0:
                            listener.setup()
                            event = listener.hooks.channel_events.pop()
                            if not event.is_set():
                                event.set()       

            except Exception as e:
                response.complete = time.monotonic()
                response.error = str(e)

                self.pool.connections.append(HTTPConnection(reset_connection=self.pool.reset_connections))

            self.active -= 1
            if self.waiter and self.active <= self.pool.size:

                try:
                    self.waiter.set_result(None)
                    self.waiter = None

                except asyncio.InvalidStateError:
                    self.waiter = None

            return response

    async def close(self):
        if self.closed is False:
            await self.pool.close()
            self.closed = True



import os
import json
from typing import Dict, Any
from hedra.logging import (
    HedraLogger,
    LoggerTypes,
    logging_manager
)


def about_project(
    path: str
):
    logging_manager.disable(
        LoggerTypes.HEDRA, 
        LoggerTypes.DISTRIBUTED,
        LoggerTypes.FILESYSTEM,
        LoggerTypes.DISTRIBUTED_FILESYSTEM
    )

    logger = HedraLogger()
    logger.initialize()
    logging_manager.logfiles_directory = os.getcwd()

    hedra_config_filepath = os.path.join(
        path,
        '.hedra.json'
    )

    hedra_config = {}
    if os.path.exists(hedra_config_filepath):
        with open(hedra_config_filepath, 'r') as config_file:
            hedra_config: Dict[str, Any] = json.load(config_file)

    else:
        logger.console.sync.error(f'No project found at path - {path}\n')
        exit(0)

    project_name = hedra_config.get('name')
    logger.console.sync.info(f'Project - {project_name}')

    logger.console.sync.info(f'\nGraphs:')
    hedra_graphs = hedra_config.get('graphs', {})

    if len(hedra_graphs) > 0:
        for graph_name, graph_path in hedra_graphs.items():
            logger.console.sync.info(f' - {graph_name} at {graph_path}')

    else:
        logger.console.sync.info(' - No graphs registered')

    logger.console.sync.info(f'\nPlugins:')
    hedra_plugins = hedra_config.get('plugins', {})

    if len(hedra_plugins) > 0:
        for plugin_name, plugin_path in hedra_plugins.items():
            logger.console.sync.info(f' - {plugin_name} at {plugin_path}')

    else:
        logger.console.sync.info(' - No plugins registered')


    logger.console.sync.info(f'\nCore Options:')
    hedra_options = hedra_config.get('core', {})
    
    if len(hedra_options) > 0:
        for opttion_name, option_value in hedra_options.items():
            logger.console.sync.info(f' - {opttion_name} set as {option_value}')

    else:
        logger.console.sync.info(' - No options set')


    logger.console.sync.info('')


import os
import uuid
from pathlib import Path
from typing import List
from git import RemoteReference
from git.repo import Repo
from git.remote import Remote
from hedra.logging import HedraLogger
from .config import RepoConfig


class RepoAction:

    def __init__(self, config: RepoConfig, discovered_files: List[str]) -> None:
        self.action_id = str(uuid.uuid4())
        self.config = config
        self.repo: Repo = None
        self.remote: Remote = None
        self.branch = None
        self.git = None
        self.discovered_files = discovered_files
        self.logger = HedraLogger()
        self.logger.initialize()

    def _pull_from_remote(self):
        self.logger.hedra.sync.debug(f'GitAction: {self.action_id} - Cloning from remote - {self.config.uri} - to path - {self.config.path}')
        self.repo = Repo.clone_from(self.config.uri, self.config.path)
        self.branch = self.repo.create_head(self.config.branch)

        self.logger.hedra.sync.debug(f'GitAction: {self.action_id} -Setting branch - {self.config.branch}')

        self.remote = self.repo.remote(name=self.config.remote)

        self.logger.hedra.sync.debug(f'GitAction: {self.action_id} -Setting remote - {self.config.remote}')

        self.git = self.repo.git

    def _setup(self):
        self.logger.hedra.sync.debug(f'GitAction: {self.action_id} - Re-initializing repo at - {self.config.path}')
        self.repo = Repo(self.config.path)

        self.logger.hedra.sync.debug(f'GitAction: {self.action_id} -Setting remote - {self.config.remote}')
        self.remote = Remote(self.repo, self.config.remote)

        self.logger.hedra.sync.debug(f'GitAction: {self.action_id} -Setting branch - {self.config.branch}')
        self.branch = self.repo.create_head(self.config.branch)
        self.git = self.repo.git

    def _checkout(self):

        if self.branch in self.repo.branches and self.branch.name != self.repo.head.name:
            self.logger.hedra.sync.debug(f'GitAction: {self.action_id} - Checking out existing branch - {self.config.branch}')
            self.branch.checkout()

        else:
            self.logger.hedra.sync.debug(f'GitAction: {self.action_id} - Creating new branch - {self.config.branch}')
            self.branch = self.repo.create_head(self.config.branch)

        self.repo.head.reference = self.branch
        
        self.logger.hedra.sync.debug(f'GitAction: {self.action_id} - Setting remote - {self.config.remote} - to track from branch - {self.config.branch}')

        remote_reference = RemoteReference(
            self.repo, 
            f"refs/remotes/{self.config.remote}/{self.branch.name}"
        )

        self.repo.head.reference.set_tracking_branch(remote_reference).checkout()
        self.logger.hedra.sync.debug(f'GitAction: {self.action_id} - Checkout complete')

    def _update_ignore(self, force_create=False):

        self.logger.hedra.sync.debug(f'GitAction: {self.action_id} - Updating .gitignore at {self.config.path}/.gitignore')

        license_file = os.path.join(self.config.path, 'LICENSE')
        readme_path = os.path.join(self.config.path, 'README.md')

        valid_files = [
            license_file,
            readme_path,
            *self.discovered_files
        ]

        gitignore_path = f'{self.config.path}/.gitignore'

        existing_ignore_files = []
        
        if os.path.exists(gitignore_path) or force_create:
            self.logger.hedra.sync.debug(f'GitAction: {self.action_id} - Creating or updating .gitignore')

            if os.path.exists(gitignore_path):
                self.logger.hedra.sync.debug(f'GitAction: {self.action_id} - Reading existing .gitignore to check for already ignored items')
                with open(gitignore_path, 'r') as hedra_gitignore:
                    existing_ignore_files.extend([
                        existing_ignore_file.strip('\n') for existing_ignore_file in hedra_gitignore.readlines()
                    ])

            with open(gitignore_path, 'a+') as hedra_gitignore:
                
                filter_files: List[str] = ['**/.hedra.json']
                candidate_filter_files = [
                    str(path.resolve()) for path in Path(self.config.path).rglob('*') if '.git' not in str(path.resolve())
                ]

                existing_ignore_files.extend(
                    self.repo.ignored(candidate_filter_files)
                )
                
                for candidate_filter_file in candidate_filter_files:

                    candidate_filter_filepath = str(Path(candidate_filter_file).resolve())

                    candidate_relative_path = os.path.relpath(candidate_filter_filepath, self.config.path)
                    
                    valid_ignore_candidate = candidate_filter_filepath not in valid_files
                    not_already_ignored = candidate_filter_filepath not in existing_ignore_files
                    not_directory = os.path.isdir(candidate_filter_filepath) is False

                    if valid_ignore_candidate and not_already_ignored and not_directory:
                        filter_files.append(candidate_relative_path)

                for ignore_option in self.config.ignore_options:
                    if ignore_option not in existing_ignore_files:
                        filter_files.append(ignore_option)

                filter_files_data = '\n'.join([
                    filepath for filepath in filter_files
                ])

                self.logger.hedra.sync.debug(f'GitAction: {self.action_id} - Adding {len(filter_files)} new files to .gitignore')

                if len(filter_files) > 0:
                    hedra_gitignore.writelines(f'\n{filter_files_data}\n')
                    
            self.repo.index.add('.gitignore')
            self.logger.hedra.sync.debug(f'GitAction: {self.action_id} - Update of .gitignore at - {self.config.path}/.gitignore - complete')



import asyncio
import functools
import uuid
import psutil
from typing import List
from concurrent.futures import ThreadPoolExecutor
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet
from .aws_timestream_config import AWSTimestreamConfig
from .aws_timestream_record import AWSTimestreamRecord
from .aws_timestream_error_record import AWSTimestreamErrorRecord

try:

    import boto3
    has_connector = True
except Exception:
    has_connector = False


class AWSTimestream:

    def __init__(self, config: AWSTimestreamConfig) -> None:
        self.aws_access_key_id = config.aws_access_key_id
        self.aws_secret_access_key = config.aws_secret_access_key
        self.region_name = config.region_name

        self.database_name = config.database_name
        self.events_table_name = config.events_table
        self.metrics_table_name = config.metrics_table
        self.stage_metrics_table_name = 'stage_metrics'
        self.errors_table_name = 'stage_errors'

        self.retention_options = config.retention_options
        self.session_uuid = str(uuid.uuid4())

        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        self.client = None
        self._loop = asyncio.get_event_loop()
        self.metadata_string: str = None

        self.logger = HedraLogger()
        self.logger.initialize()

    async def connect(self):

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Opening session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opening amd authorizing connection to AWS - Region: {self.region_name}')

        self.client = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                boto3.client,
                'timestream-write',
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                region_name=self.region_name
            )
        )

        try:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating table - Database: {self.database_name} - if not exists')

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_database,
                    DatabaseName=self.database_name
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created table - Database: {self.database_name} - if not exists')
        
        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Skipping creation of table - Database: {self.database_name} - if not exists')
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Successfully opened connection to AWS - Region: {self.region_name}')

    async def submit_events(self, events: List[BaseEvent]):

        try:
            
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating table - Database: {self.database_name} - Table: {self.events_table_name} - if not exists')

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_table,
                    DatabaseName=self.database_name,
                    TableName=self.events_table_name,
                    RetentionProperties=self.retention_options
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created table - Database: {self.database_name} - Table: {self.events_table_name} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Skipping creation of table - Database: {self.database_name} - Table: {self.events_table_name} - if not exists')

        records = []
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events - Database: {self.database_name} - Table: {self.events_table_name} - if not exists')

        for event in events:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Event - {event.name}:{event.event_id}')

            for field, value in event.record.items():
                timestream_record = AWSTimestreamRecord(
                    'event',
                    event.name,
                    field,
                    value,
                    self.session_uuid
                )

                records.append(timestream_record.to_dict())

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.write_records,
                DatabaseName=self.database_name,
                TableName=self.events_table_name,
                Records=records,
                CommonAttributes={}
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events - Database: {self.database_name} - Table: {self.events_table_name} - if not exists')

    async def submit_common(self, metrics: List[MetricsSet]):
        try:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating table - Database: {self.database_name} - Table: {self.metrics_table_name} - if not exists')

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_table,
                    DatabaseName=self.database_name,
                    TableName=self.stage_metrics_table_name,
                    RetentionProperties=self.retention_options
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created table - Database: {self.database_name} - Table: {self.metrics_table_name} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Skipping creation of table - Database: {self.database_name} - Table: {self.metrics_table_name} - if not exists')

        records = []
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics - Database: {self.database_name} - Table: {self.events_table_name} - if not exists')

        for metrics_set in metrics:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
 
            for field, value in metrics_set.common_stats.items():
                timestream_record = AWSTimestreamRecord(
                    'stage_metrics',
                    metrics_set.name,
                    metrics_set.stage,
                    'common',
                    field,
                    value,
                    self.session_uuid,
                )

                records.append(timestream_record.to_dict())
                    
        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.write_records,
                DatabaseName=self.database_name,
                TableName=self.stage_metrics_table_name,
                Records=records,
                CommonAttributes={}
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics - Database: {self.database_name} - Table: {self.events_table_name} - if not exists')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        try:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating table - Database: {self.database_name} - Table: {self.metrics_table_name} - if not exists')

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_table,
                    DatabaseName=self.database_name,
                    TableName=self.metrics_table_name,
                    RetentionProperties=self.retention_options
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created table - Database: {self.database_name} - Table: {self.metrics_table_name} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Skipping creation of table - Database: {self.database_name} - Table: {self.metrics_table_name} - if not exists')

        records = []
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics - Database: {self.database_name} - Table: {self.events_table_name} - if not exists')

        for metrics_set in metrics:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            metrics_groups = {**metrics_set.groups, **metrics_set.custom_metrics}

            for group_name, group in metrics_groups.items():

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Group - {group_name}')
            
                metric_result = {**group.stats, **group.custom}

                for field, value in metric_result.items():
                    timestream_record = AWSTimestreamRecord(
                        'metric',
                        metrics_set.name,
                        metrics_set.stage,
                        group_name,
                        field,
                        value,
                        self.session_uuid,
                    )

                    records.append(timestream_record.to_dict())

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.write_records,
                DatabaseName=self.database_name,
                TableName=self.metrics_table_name,
                Records=records,
                CommonAttributes={}
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics - Database: {self.database_name} - Table: {self.events_table_name} - if not exists')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping Custom Metrics submission')
        
    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        try:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating table - Database: {self.database_name} - Table: {self.errors_table_name} - if not exists')

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_table,
                    DatabaseName=self.database_name,
                    TableName=self.errors_table_name,
                    RetentionProperties=self.retention_options
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created table - Database: {self.database_name} - Table: {self.errors_table_name} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Skipping creation of table - Database: {self.database_name} - Table: {self.errors_table_name} - if not exists')

        error_records = []
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Errors Metrics - Database: {self.database_name} - Table: {self.events_table_name} - if not exists')

        for metrics_set in metrics_sets:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Errors Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                timestream_record = AWSTimestreamErrorRecord(
                    metrics_set.name,
                    metrics_set.stage,
                    error.get('message'),
                    error.get('count'),
                    self.session_uuid
                )

                error_records.append(timestream_record.to_dict())

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.write_records,
                DatabaseName=self.database_name,
                TableName=f'{self.metrics_table_name}_errors',
                Records=error_records,
                CommonAttributes={}
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Errors Metrics - Database: {self.database_name} - Table: {self.events_table_name} - if not exists')


    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')


import asyncio
import time
import uuid
from hedra.core.personas.types.default_persona import DefaultPersona
from hedra.core.engines.client.config import Config
from hedra.core.personas.types.types import PersonaTypes


class CyclicNoWaitPersona(DefaultPersona):

    def __init__(self, config: Config):
        super().__init__(config)

        self.persona_id = str(uuid.uuid4())
        self.type = PersonaTypes.NO_WAIT

    async def generator(self, total_time):
        elapsed = 0
        idx = 0

        start = time.monotonic()
        while elapsed < total_time:
            yield idx%self.actions_count
            
            await asyncio.sleep(0)
            elapsed = time.monotonic() - start
            idx += 1
            


import os
from pydantic import BaseModel
from typing import Optional
from hedra.reporting.types.common.types import ReporterTypes


class CSVConfig(BaseModel):
    events_filepath: Optional[str]=f'{os.getcwd()}/events.csv'
    metrics_filepath: str=f'{os.getcwd()}/metrics.csv'
    reporter_type: ReporterTypes=ReporterTypes.CSV


import asyncio
from .bootstrap_client import BoostrapClient


class BootstrapManager:

    def __init__(self, config) -> None:
        self.client = BoostrapClient(config)
        self.use_bootstrap = self.client.use_bootstrap

    async def discovered(self):
        return list(self.client.registry.discovered.values())



from typing import Any
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.graphs.transitions.exceptions import IdleTranstionError
from hedra.core.graphs.transitions.exceptions import StageExecutionError
from hedra.logging import HedraLogger


async def invalid_idle_transition(current_stage: Stage, next_stage: Stage):
    return IdleTranstionError(next_stage), StageTypes.ERROR


async def idle_to_validate_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    try:
        
        await logger.spinner.system.debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')
        await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')
        await current_stage.run()

        next_stage.context = current_stage.context

    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    else:
        return None, StageTypes.VALIDATE

async def idle_to_wait_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    try:

        await logger.spinner.system.debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')
        await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - NoOp transition from {current_stage.name} to {next_stage.name}')
        await current_stage.run()

        next_stage.context = current_stage.context

    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    else:
        return None, StageTypes.WAIT


from typing import List, Type
from .connection import CustomConnection


class CustomPool:

    __slots__ = (
        'size',
        'connections',
        'reset_connections',
        'create_connection_type'
    )

    def __init__(self, create_connection_type: Type[CustomConnection], size: int, reset_connections: bool = False) -> None:
        self.size = size
        self.connections: List[CustomConnection] = []
        self.reset_connections = reset_connections
        self.create_connection_type: CustomConnection = create_connection_type

    def create_pool(self) -> None:
        for _ in range(self.size):
            self.connections.append(
                self.create_connection_type(self.reset_connections)
            )

    async def close(self):
        for connection in self.connections:
            await connection.close()


from .headers_sent_event import HeadersSent


class RequestSent(HeadersSent):
    event_type='REQUEST_SENT'
    """
    The _RequestSent event is fired whenever request headers are sent
    on a stream.

    This is an internal event, used to determine validation steps on
    outgoing header blocks.
    """
    pass


from .generated_engine_plugin import (
    CustomEngine
)

from .generated_optimizer_plugin import (
    CustomOptimizer,
)

from .generated_persona_plugin import (
    CustomPersona
)

from .generated_reporter_plugin import (
    CustomReporter
)


from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.http2.result import HTTP2Result
from .action import GraphQLHTTP2Action


class GraphQLHTTP2Result(HTTP2Result):

    def __init__(self, action: GraphQLHTTP2Action, error: Exception = None) -> None:
        super().__init__(action, error)
        self.type = RequestTypes.GRAPHQL_HTTP2


from typing import Any, Coroutine


class CustomConnection:

    __slots__ = (
        "security_context",
        "reset_connection",
        "connected",
        "pending",
        "make_connection",
        "execute",
        "close",
    )

    def __init__(
        self, 
        security_context: Any, 
        reset_connection: bool=False,
        on_connect: Coroutine=None,
        on_execute: Coroutine=None,
        on_close: Coroutine=None
    ) -> None:
        self.security_context = security_context
        self.reset_connection = reset_connection
        self.connected = False
        self.pending = 0
        
        self.make_connection = on_connect
        self.execute = on_execute
        self.close = on_close



import asyncio
import traceback
from typing import Dict, List, Union
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.types.stage_states import StageStates
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.graphs.transitions.exceptions import (
    StageExecutionError,
    StageTimeoutError
)
from hedra.core.engines.types.common.base_result import BaseResult
from hedra.logging import HedraLogger


async def execute_transition(current_stage: Stage, next_stage: Stage, logger: HedraLogger):

    execute_stages = current_stage.context.stages.get(StageTypes.EXECUTE)

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - Executing transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Executing transition from {current_stage.name} to {next_stage.name}')

    total_concurrent_execute_stages = []
    for generation_stage_name in current_stage.generation_stage_names:
        stage = execute_stages.get(generation_stage_name)
        
        if stage is not None:
            total_concurrent_execute_stages.append(stage)

    current_stage.total_concurrent_execute_stages = len(total_concurrent_execute_stages)
    
    if current_stage.timeout:
        execution_results: Dict[str, Union[List[BaseResult], int, float]] = await asyncio.wait_for(current_stage.run(), timeout=current_stage.timeout)

    else:
        execution_results: Dict[str, Union[List[BaseResult], int, float]] = await current_stage.run()

    if current_stage.context.results.get(current_stage.name) is None:
        current_stage.context.results[current_stage.name] = {
            'results': execution_results.get('results'),
            'total_results': execution_results.get('total_results'),
            'total_elapsed': execution_results.get('total_elapsed')
        }
    
    else:
        current_stage.context.results[current_stage.name].update({
            'results': execution_results.get('results'),
            'total_results': execution_results.get('total_results'),
            'total_elapsed': execution_results.get('total_elapsed')
        })

    current_stage.context.results_stages.append(current_stage)
    current_stage.context.visited.append(current_stage.name)

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - Completed transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Completed transition from {current_stage.name} to {next_stage.name}')
    
    next_stage.context = current_stage.context

    return current_stage, next_stage


async def execute_to_setup_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    try:
        valid_states = [
            StageStates.SETUP,
            StageStates.OPTIMIZED
        ]

        if current_stage.state in valid_states:
            current_stage.state = StageStates.EXECUTING

            current_stage, next_stage = await execute_transition(current_stage, next_stage, logger)
            execute_stages = list(current_stage.context.stages.get(StageTypes.EXECUTE).values())
            visited = current_stage.context.visited

            for stage in execute_stages:
                if stage.name not in visited and stage.state == StageStates.SETUP:
                    stage.state = StageStates.INITIALIZED

            current_stage.state = StageStates.EXECUTED

        else:
            await logger.spinner.system.debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')
            await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')

    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR

    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    current_stage = None

    return None, StageTypes.SETUP


async def execute_to_execute_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    try:
        valid_states = [
            StageStates.SETUP,
            StageStates.OPTIMIZED
        ]

        if current_stage.state in valid_states:
            current_stage.state = StageStates.EXECUTING

            current_stage, next_stage = await execute_transition(current_stage, next_stage, logger)

            current_stage.state = StageStates.EXECUTED

        else:
            await logger.spinner.system.debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')
            await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')

    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR

    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    current_stage = None

    return None, StageTypes.EXECUTE


async def execute_to_optimize_transition(current_stage: Stage, next_stage: Stage):
    
    logger = HedraLogger()
    logger.initialize()

    try:
        valid_states = [
            StageStates.SETUP,
            StageStates.OPTIMIZED
        ]

        if current_stage.state in valid_states:
            current_stage.state = StageStates.EXECUTING

            current_stage, next_stage = await execute_transition(current_stage, next_stage, logger)

            current_stage.state = StageStates.EXECUTED

        else:
            await logger.spinner.system.debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')
            await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR
    
    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    current_stage = None

    return None, StageTypes.OPTIMIZE


async def execute_to_teardown_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    try:

        valid_states = [
            StageStates.SETUP,
            StageStates.OPTIMIZED
        ]

        if current_stage.state in valid_states:
            current_stage.state = StageStates.EXECUTING

            current_stage, next_stage = await execute_transition(current_stage, next_stage, logger)

            current_stage.state = StageStates.EXECUTED

        else:
            await logger.spinner.system.debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')
            await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR

    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    current_stage = None

    return None, StageTypes.TEARDOWN


async def execute_to_analyze_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    try:
        
        valid_states = [
            StageStates.SETUP,
            StageStates.OPTIMIZED
        ]

        if current_stage.state in valid_states:
            current_stage.state = StageStates.EXECUTING

            current_stage, next_stage = await execute_transition(current_stage, next_stage, logger)

            current_stage.state = StageStates.EXECUTED

        else:
            await logger.spinner.system.debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')
            await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR

    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    current_stage = None

    return None, StageTypes.ANALYZE


async def execute_to_checkpoint_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    try:

        valid_states = [
            StageStates.SETUP,
            StageStates.OPTIMIZED
        ]

        if current_stage.state in valid_states:
            current_stage.state = StageStates.EXECUTING

            current_stage, next_stage = await execute_transition(current_stage, next_stage, logger)
            next_stage.data = current_stage.context.results

            next_stage.previous_stage = current_stage.name

            current_stage.state = StageStates.EXECUTED

        else:
            await logger.spinner.system.debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')
            await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR

    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    current_stage = None

    return None, StageTypes.CHECKPOINT


async def execute_to_wait_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    try:

        valid_states = [
            StageStates.SETUP,
            StageStates.OPTIMIZED
        ]

        if current_stage.state in valid_states:
            current_stage.state = StageStates.EXECUTING

            current_stage, next_stage = await execute_transition(current_stage, next_stage, logger)
            next_stage.data = current_stage.context.results[current_stage.name]

            next_stage.previous_stage = current_stage.name

            current_stage.state = StageStates.EXECUTED

        else:
            await logger.spinner.system.debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')
            await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR

    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    current_stage = None

    return None, StageTypes.WAIT



import asyncio
import sys

from io import (
    FileIO,
    TextIOBase,
    BufferedReader,
    BufferedWriter,
    BufferedRandom,
    BufferedIOBase,
)
from functools import partial, singledispatch

from .binary import (
    AsyncBufferedIOBase,
    AsyncBufferedReader,
    AsyncFileIO,
    AsyncIndirectBufferedIOBase
)
from .text import AsyncTextIOWrapper, AsyncTextIndirectIOWrapper


sync_open = open

__all__ = (
    "open",
    "stdin",
    "stdout",
    "stderr",
    "stdin_bytes",
    "stdout_bytes",
    "stderr_bytes",
)


async def open(
    file,
    mode="r",
    buffering=-1,
    encoding=None,
    errors=None,
    newline=None,
    closefd=True,
    opener=None,
    *,
    loop=None,
    executor=None
) -> AsyncTextIOWrapper:

    return await _open(
        file,
        mode=mode,
        buffering=buffering,
        encoding=encoding,
        errors=errors,
        newline=newline,
        closefd=closefd,
        opener=opener,
        loop=loop,
        executor=executor,
    )


async def _open(
    file,
    mode="r",
    buffering=-1,
    encoding=None,
    errors=None,
    newline=None,
    closefd=True,
    opener=None,
    *,
    loop=None,
    executor=None
):
    """Open an asyncio file."""
    if loop is None:
        loop = asyncio.get_event_loop()
    cb = partial(
        sync_open,
        file,
        mode=mode,
        buffering=buffering,
        encoding=encoding,
        errors=errors,
        newline=newline,
        closefd=closefd,
        opener=opener,
    )
    async_file = await loop.run_in_executor(executor, cb)

    return wrap(async_file, loop=loop, executor=executor)


@singledispatch
def wrap(file, *, loop=None, executor=None):
    raise TypeError("Unsupported io type: {}.".format(file))


@wrap.register(TextIOBase)
def _(file, *, loop=None, executor=None):
    return AsyncTextIOWrapper(file, loop=loop, executor=executor)


@wrap.register(BufferedWriter)
@wrap.register(BufferedIOBase)
def _(file, *, loop=None, executor=None):
    return AsyncBufferedIOBase(file, loop=loop, executor=executor)


@wrap.register(BufferedReader)
@wrap.register(BufferedRandom)
def _(file, *, loop=None, executor=None):
    return AsyncBufferedReader(file, loop=loop, executor=executor)


@wrap.register(FileIO)
def _(file, *, loop=None, executor=None):
    return AsyncFileIO(file, loop=loop, executor=executor)


stdin = AsyncTextIndirectIOWrapper('sys.stdin', None, None, indirect=lambda: sys.stdin)
stdout = AsyncTextIndirectIOWrapper('sys.stdout', None, None, indirect=lambda: sys.stdout)
stderr = AsyncTextIndirectIOWrapper('sys.stderr', None, None, indirect=lambda: sys.stderr)
stdin_bytes = AsyncIndirectBufferedIOBase('sys.stdin.buffer', None, None, indirect=lambda: sys.stdin.buffer)
stdout_bytes = AsyncIndirectBufferedIOBase('sys.stdout.buffer', None, None, indirect=lambda: sys.stdout.buffer)
stderr_bytes = AsyncIndirectBufferedIOBase('sys.stderr.buffer', None, None, indirect=lambda: sys.stderr.buffer)


import os
from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    CassandraConfig
)


class SubmitCassandraResultsStage(Submit):
    config=CassandraConfig(
        hosts=['127.0.0.1'],
        port=9042,
        username=os.getenv('CASSANDRA_USERNAME', ''),
        password=os.getenv('CASSANDRA_PASSWORD', ''),
        keyspace='results',
        events_table='events',
        metrics_table='metrics'
    )


import os
import json
from flask import Flask
from .services import JobsService


class JobServer:
    
    def __init__(self, config, reporter_config):
        self.app = Flask(__name__)
        self.jobs_service = JobsService(config, reporter_config)

    def get_app(self):
        return self.app

    def run(self, host=None, port=None):
        self.jobs_service.host = host
        self.jobs_service.port = port
        self.app.run(host=host, port=port)

    def setup(self):
        for endpoint_config, endpoint_handler in self.jobs_service:   
            self.app.route(
                endpoint_config.get('path'),
                methods=endpoint_config.get('methods')
            )(endpoint_handler)



import asyncio
import os
import re
import uuid
import grpc
from flask import request
from hedra.servers.job_server.config_store import ConfigStore
from hedra.runners.leader_services.proto import(
    DistributedServerStub,
    NewJobLeaderRequest
)




class JobsService:

    def __init__(self, config, reporter_config):
        self.server_id = uuid.uuid4()
        self.host = None
        self.port = None
        self.leader_ips = config.distributed_config.get(
            'leader_ips',
            os.getenv('LEADER_IPS', 'localhost')
        )

        self.leader_ports = config.distributed_config.get(
            'leader_ports',
            os.getenv('LEADER_PORTS', "6669")
        )

        leader_addresses = config.distributed_config.get(
            'leader_addresses',
            os.getenv('LEADER_ADDRESSES')
        )

        if leader_addresses:
            leader_addresses = leader_addresses.split(",")
            self.leader_addresses = leader_addresses

        else:
            self.leader_ips = self.leader_ips.split(",")
            self.leader_ports = [int(port) for port in self.leader_ports.split(",")]

            self.leader_addresses = []

            for leader_ip, leader_port in zip(self.leader_ips, self.leader_ports):
                self.leader_addresses.append(
                    f'{leader_ip}:{leader_port}'
                )

        self.leaders = {}
        logger = Logger()
        self.session_logger = logger.generate_logger('hedra')
        self.session_logger.info('Initializing server...')

        self.endpoint_configs = [
            {
                'name': 'setup_job',
                'path': '/api/hedra/jobs/setup',
                'methods': ['POST', 'PUT']
            },
            {
                'name': 'delete_job',
                'path': '/api/hedra/jobs/delete',
                'methods': ['DELETE']
            },
            {
                'name': 'start_job',
                'path': '/api/hedra/jobs/start',
                'methods': ['GET']
            },
            {
                'name': 'get_job_config',
                'path': '/api/hedra/jobs/get',
                'methods': ['GET']
            }
        ]

        self.store = ConfigStore()

    def __iter__(self):
        for endpoint_config in self.endpoint_configs:
            yield endpoint_config, getattr(self, endpoint_config.get('name'))

    async def setup_job(self):
        config_json = request.json
        job_name = self.store.add_config(config_json)
        return {
            'message': 'Config setup for job - {job_name}.'.format(job_name=job_name)
        }, 201

    async def delete_job(self):
        job_name = request.args.get('job')
        self.store.delete_config(job_name)
        return {
            'message': f'Job - {job_name} deleted.'
        }, 204

    async def start_job(self):
        job_name = request.args.get('job')
        timeout = request.args.get('timeout', 300)
        
        if job_name is None:
            return {
                'message': 'Error: No job name provided.'
            }, 400

        config, reporter_config = self.store.get(job_name)

        if config is None:
            return {
                'message': 'Error: No config found for job - {job_name}'.format(
                    job_name=job_name
                )
            }, 400

        if reporter_config is None:
            return {
                'message': 'Error: No reporter config found for job - {job_name}'.format(
                    job_name=job_name
                )
            }, 400

        response = None

        leader_jobs = []
        for leader_address in self.leader_addresses:
            channel = grpc.aio.insecure_channel(leader_address)
            self.leaders[leader_address] = DistributedServerStub(channel)

        for _ in self.leaders.values():

            new_job = NewJobLeaderRequest(
                host_address=f'{self.host}:{self.port}',
                host_id=str(self.server_id),
                job_timeout=int(timeout)
            )
            new_job.job_config.update({
                'job_name': job_name,
                'executor_config': config.executor_config,
                'actions': config.actions
            })
            

            leader_jobs += [new_job]

        leader_jobs, _ = await asyncio.wait([
            leader.CreateNewJob(job) for leader, job in zip(
                self.leaders.values(),
                leader_jobs
            )
        ])

        leader_jobs = await asyncio.gather(*leader_jobs)

        for job in leader_jobs:
            if isinstance(response, grpc.RpcError):
                return {
                    'message': f'Error: Leader at - {job.host_address} - could not spawn job. Runner is busy.'
                }, 400

        return {
            'message': 'Job started.',
            'job_name': job_name
        }, 200
                
    async def get_job_config(self):
        job_name = request.args.get('job')
        config, reporter_config = self.store.get(job_name)
        if config is None:
            return {
                'message': 'Error: Config for job - {job_name} - not found.'.format(job_name=job_name)
            }, 404

        if reporter_config is None:
            return {
                'message': 'Error: Reporter config for job - {job_name} - not found.'.format(job_name=job_name)
            }, 404

        return {
            'message': 'Successfully found config for job - {job_name}'.format(job_name=job_name),
            'config': {
                'job_name': job_name,
                'executor_config': config.executor_config,
                'distributed_config': config.distributed_config,
                'reporter_config': reporter_config.reporter_config,
                'streams_config': reporter_config.stream_config
            }
        }, 200



import os
from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    AWSLambdaConfig
)


class SubmitAWSLambdaResultsStage(Submit):
    config=AWSLambdaConfig(
        aws_access_key_id=os.getenv('AWS_ACCESS_KEY_ID', ''),
        aws_secret_access_key=os.getenv('AWS_SECRET_ACCESS_KEY', ''),
        region_name=os.getenv("AWS_REGION_NAME", ''),
        events_lambda='events',
        metrics_lambda='metrics'
    )


import asyncio
import uuid
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
import uuid
from datetime import datetime
from typing import List

import psutil
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet

try:
    from google.cloud import bigtable
    from google.auth import load_credentials_from_file
    from .bigtable_config import BigTableConfig
    has_connector = True

except Exception:
    bigtable = None
    Credentials = None
    BigTableConfig = None
    has_connector = False


class BigTable:

    def __init__(self, config: BigTableConfig) -> None:
        self.service_account_json_path = config.service_account_json_path
        self.instance_id = config.instance_id
        self.events_table_id = config.events_table
        self.metrics_table_id = config.metrics_table
        self.shared_metrics_table_id = f'{self.metrics_table_id}_metrics'
        self.custom_metrics_table_ids = {}
        self.errors_table_id = f'{self.metrics_table_id}_errors'

        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        self._events_column_family_id = f'{self.events_table_id}_columns'
        self._metrics_column_family_id = f'{self.metrics_table_id}_columns'
        self._shared_metrics_column_family_id = 'stage_metrics_columns'
        self._custom_metrics_column_family_ids = {}
        self._errors_column_family_id = f'{self.metrics_table_id}_errors_columns'

        self.instance = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self._events_table = None
        self._stage_metrics_table = None
        self._custom_metrics_tables = {}
        self._metrics_table = None
        self._errors_table = None

        self._events_table_columns = None
        self._stage_metrics_table_columns = None
        self._custom_metrics_table_columns = {}
        self._metrics_table_columns = None
        self._errors_table_columns = None

        self.credentials = None
        self.client = None
        self._loop = asyncio.get_event_loop()

    async def connect(self):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opening amd authorizing connection to Google Cloud - Loading account config from - {self.service_account_json_path}')
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Opening session - {self.session_uuid}')

        credentials, project_id = load_credentials_from_file(self.service_account_json_path)
        self.client = bigtable.Client(
            project=project_id,
            credentials=credentials,
            admin=True
        )
        self.instance = self.client.instance(self.instance_id)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opened connection to Google Cloud - Created Client Instance - ID:{self.instance_id}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opened connection to Google Cloud - Loaded account config from - {self.service_account_json_path}')

    async def submit_events(self, events: List[BaseEvent]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Table {self.events_table_id}')
        self._events_table = self.instance.table(self.events_table_id)
        
        try:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Events Column Family - {self._events_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._events_table.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Events Column Family - {self._events_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Events Column Family - {self._events_column_family_id} - if not exists')

        self._events_table_columns = self._events_table.column_family(
            self._events_column_family_id
        )

        try:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Events Column for Column Family - {self._events_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._events_table_columns.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Events Column for Column Family - {self._events_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Events Column for Column Family - {self._events_column_family_id} - if not exists')

        rows = []
        for event in events:
            row_key = f'{event.name}_{str(uuid.uuid4())}'
            row = self._events_table.direct_row(row_key)

            for field, value in event.record.items():
                if not isinstance(value, bytes):
                    value = f'{value}'.encode()

                row.set_cell(
                    self._events_column_family_id,
                    field,
                    value,
                    timestamp=datetime.now()
                )

            rows.append(row)

        await self._loop.run_in_executor(
            self._executor,
            self._events_table.mutate_rows,
            rows
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Table {self.events_table_id}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Table {self.shared_metrics_table_id}')

        stage_metrics_table = self.instance.table(self.shared_metrics_table_id)
            
        try:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Shared Metrics Column Family - {self._shared_metrics_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                stage_metrics_table.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Shared Metrics Column Family - {self._shared_metrics_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Shared Metrics Column Family - {self._shared_metrics_column_family_id} - if not exists')

        self._metrics_table_columns = stage_metrics_table.column_family(
            self._shared_metrics_column_family_id
        )

        
        try:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Shared Metrics Column for Column Family - {self._shared_metrics_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._metrics_table_columns.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Shared Metrics Column for Column Family - {self._shared_metrics_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Shared Metrics Column for Column Family - {self._shared_metrics_column_family_id} - if not exists')


        rows = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            row_key = f'{self.shared_metrics_table_id}_{str(uuid.uuid4())}'
            row = stage_metrics_table.direct_row(row_key)

            stage_metrics_record = {
                'name': metrics_set.name,
                'stage': metrics_set.stage,
                'group': 'common',
                **metrics_set.common_stats
            }

            for field, value in stage_metrics_record.items():
                if not isinstance(value, bytes):
                    value = f'{value}'.encode()

                row.set_cell(
                    self._metrics_column_family_id,
                    field,
                    value,
                    timestamp=datetime.now()
                )

            rows.append(row)

        await self._loop.run_in_executor(
            self._executor,
            stage_metrics_table.mutate_rows,
            rows
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Table {self.shared_metrics_table_id}')
    
    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Table {self.metrics_table_id}')

        metrics_table = self.instance.table(self.metrics_table_id)
            
        try:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Metrics Column Family - {self._metrics_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                metrics_table.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Metrics Column Family - {self._metrics_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Metrics Column Family - {self._metrics_column_family_id} - if not exists')

        self._metrics_table_columns = metrics_table.column_family(
            self._metrics_column_family_id
        )
        
        try:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Metrics Column for Column Family - {self._metrics_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._metrics_table_columns.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Metrics Column for Column Family - {self._metrics_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Metrics Column for Column Family - {self._metrics_column_family_id} - if not exists')
        
        rows = []
        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for group_name, group in metrics_set.groups.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Group - {group_name}:{group.metrics_group_id}')

                row_key = f'{self.metrics_table_id}_{str(uuid.uuid4())}'
                row = metrics_table.direct_row(row_key)

                record = {
                    'group': group_name,
                    **group.record
                }

                for field, value in record.items():
                    if not isinstance(value, bytes):
                        value = f'{value}'.encode()

                    row.set_cell(
                        self._metrics_column_family_id,
                        field,
                        value,
                        timestamp=datetime.now()
                    )

                rows.append(row)

        await self._loop.run_in_executor(
            self._executor,
            metrics_table.mutate_rows,
            rows
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Table {self.metrics_table_id}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to Table {self.metrics_table_id}')

        rows = defaultdict(list)
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for custom_group_name, custom_group in metrics_set.custom_metrics.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Group - {custom_group_name}')

                custom_metrics_table_id = f'{custom_group_name}_metrics'
                custom_metrics_table = self.instance.table(custom_metrics_table_id)

                try:
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Custom Metrics Column Family - {self._custom_metrics_column_family_ids.get(custom_group_name)} - if not exists')
                    await self._loop.run_in_executor(
                        self._executor,
                        custom_metrics_table.create
                    )

                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Custom Metrics Column Family - {self._custom_metrics_column_family_ids.get(custom_group_name)} - if not exists')
                
                except Exception:
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Custom Metrics Column Family - {self._custom_metrics_column_family_ids.get(custom_group_name)} - if not exists')

                self._custom_metrics_tables[custom_metrics_table_id] = custom_metrics_table

                custom_metrics_table_columns = custom_metrics_table.column_family(
                    custom_metrics_table_id
                )

                try:
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Custom Metrics Column for Column Family - {self._custom_metrics_column_family_ids.get(custom_group_name)} - if not exists')
                    await self._loop.run_in_executor(
                        self._executor,
                        custom_metrics_table_columns.create
                    )

                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Custom Metrics Column for Column Family - {self._custom_metrics_column_family_ids.get(custom_group_name)} - if not exists')

                except Exception:
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Custom Metrics Column for Column Family - {self._custom_metrics_column_family_ids.get(custom_group_name)} - if not exists')

                row_key = f'{custom_metrics_table_id}_{str(uuid.uuid4())}'
                custom_metrics_table_row = custom_metrics_table.direct_row(row_key)

                custom_metrics_row_data = {
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': custom_group_name,
                    **custom_group
                }

                for field, value in custom_metrics_row_data.items():
                    custom_metrics_table_row.set_cell(
                        field,
                        f'{value}'.encode()
                    )
                
                rows[custom_metrics_table_id].append(custom_metrics_table_row)

        for table_id, table_rows in rows.items():
            table = self._custom_metrics_tables.get(table_id)
            await self._loop.run_in_executor(
                self._executor,
                table.mutate_rows,
                table_rows
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to Table {self.metrics_table_id}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to Table {self.errors_table_id}')

        errors_table = self.instance.table(self.errors_table_id)
            
        try:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Error Metrics Column Family - {self._errors_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                errors_table.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Error Metrics Column Family - {self._errors_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Error Metrics Column Family - {self._errors_column_family_id} - if not exists')

        self._errors_table = errors_table

        self._errors_table_columns = errors_table.column_family(
            self._errors_column_family_id
        )

        try:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Error Column for Column Family - {self._errors_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._metrics_table_columns.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Error Column for Column Family - {self._errors_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Error Column for Column Family - {self._errors_column_family_id} - if not exists')
        
        rows = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            
            for error in metrics_set.errors:

                row_key = f'{self.errors_table_id}_{str(uuid.uuid4())}'
                errors_row = errors_table.direct_row(row_key)

                error_message = error.get('message')
                error_count = error.get('count')

                errors_row.set_cell(
                    self._metrics_column_family_id,
                    'error_type',
                    f'{error_message}'.encode()
                )

                errors_row.set_cell(
                    self._metrics_column_family_id,
                    'error_count',
                    f'{error_count}'.encode()
                )

                rows.append(errors_row)
                
        await self._loop.run_in_executor(
            self._executor,
            errors_table.mutate_rows,
            rows
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to Table {self.errors_table_id}')

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
        await self._loop.run_in_executor(
            self._executor,
            self.client.close
        )


        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Session Closed - {self.session_uuid}')
        


import inspect
from collections import defaultdict
from types import MethodType
from typing import List, Dict
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.graphs.hooks.registry.registrar import registrar
from hedra.core.graphs.hooks.registry.registry_types.hook import Hook
from hedra.core.graphs.stages.validate.exceptions import (
    HookValidationError,
    ReservedMethodError
)
from hedra.logging import HedraLogger


class StageValidator:

    def __init__(self, stage: Stage, metadata_string: str) -> None:

        self.logger = HedraLogger()
        self.logger.initialize()
        self.metadata_string: str = metadata_string
        self.stage: Stage = stage
        self.hooks_by_name: Dict[str, Hook] = defaultdict(dict)
        self.hooks: Dict[HookType, List[Hook]] = {
            hook_type: [] for hook_type in HookType
        }
                
    async def validate_stage_internal_hooks(self):            
        

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Checking internal Hooks for stage - {self.stage.name} - of type - {self.stage.stage_type.name}')
        
        for reserved_hook_name in self.stage.internal_hooks:

            try:
                
                base_stage_name = self.stage.__class__.__base__.__name__
                hook = registrar.reserved[base_stage_name].get(reserved_hook_name)

                assert hasattr(self.stage, reserved_hook_name)
                assert hasattr(self.stage.__class__.__base__, reserved_hook_name)
                assert isinstance(hook, Hook)
                assert hook.hook_type == HookType.INTERNAL
                
                internal_hook = getattr(self.stage, hook.shortname)
                assert inspect.getsource(internal_hook) == inspect.getsource(hook.call)

            except AssertionError:
                raise ReservedMethodError(self.stage, reserved_hook_name)
            
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Found internal Hook - {hook.name}:{hook.hook_id}- for stage - {base_stage_name}')

    async def validate_stage_hooks(self):
        methods = inspect.getmembers(self.stage, predicate=inspect.ismethod)

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Checking Hooks for stage - {self.stage.name} - of type - {self.stage.stage_type}')

        for _, method in methods:

            method_name = method.__qualname__
            hook: Hook = registrar.all.get(method_name)
            
            if hook:
                
                base_stage_name = self.stage.__class__.__base__.__name__
                valid_hook_types_string = ', '.join([hook_type.name for hook_type in self.stage.accepted_hook_types])

                try:
                
                    assert isinstance(hook.name, str), "Hook name must be string."
                    assert isinstance(hook.hook_id, str), "Hook ID must be string."
                    assert isinstance(hook.shortname, str), f"Hook shortname for hook {hook.name}:{hook.hook_id} must be a valid string."
                    assert isinstance(hook._call, MethodType), f"Hook {hook.name}:{hook.hook_id} call is not a vaid method. All hooks must be a method."
                    assert inspect.iscoroutinefunction(hook.call), f"Hook {hook.name}:{hook.hook_id} call is not a vaid coroutine. All hook methods must be a async."
                    assert hook.hook_type in self.stage.accepted_hook_types, f"Hook {hook.name}:{hook.hook_id} is not a valid hook type for stage {self.stage.name}. Valid hook types are {valid_hook_types_string}."
                
                except AssertionError as hook_validation_error:
                    raise HookValidationError(hook.stage_instance, str(hook_validation_error))

                self.hooks[hook.hook_type].append(hook)

                self.hooks_by_name[hook.name] = hook

                await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Found Hook - {hook.name}:{hook.hook_id}- of type - {hook.hook_type.name.capitalize()} - for stage - {base_stage_name}')



import uuid
from hedra.logging import HedraLogger
from typing import Any, Callable, Awaitable
from hedra.core.graphs.hooks.hook_types.hook_type import HookType


class Hook:

    def __init__(
        self, 
        name: str, 
        shortname: str,
        call: Callable[..., Awaitable[Any]], 
        stage: str = None,
        hook_type=HookType.ACTION
    ) -> None:
        self.hook_id = str(uuid.uuid4())
        self.name = name
        self.shortname = shortname
        self._call: Callable[..., Awaitable[Any]] = call
        self.stage = stage
        self.hook_type = hook_type
        self.stage_instance: Any = None
